@@ -8,16 +8,17 @@ use reqwest::header::HeaderMap;
 use reqwest_eventsource::{Event, EventSource, RequestBuilderExt};
 use serde_json::Value;
 use std::fmt::Debug;
+use tensorzero_core::config::ConfigFileGlob;
 use tensorzero_core::endpoints::datasets::StaleDatasetResponse;
 pub use tensorzero_core::endpoints::optimization::LaunchOptimizationParams;
 pub use tensorzero_core::endpoints::optimization::LaunchOptimizationWorkflowParams;
 use tensorzero_core::endpoints::optimization::{launch_optimization, launch_optimization_workflow};
 use tensorzero_core::endpoints::stored_inference::render_samples;
-use tensorzero_core::howdy::setup_howdy;
 pub use tensorzero_core::optimization::{OptimizationJobHandle, OptimizationJobInfo};
 use tensorzero_core::stored_inference::StoredSample;
 use tensorzero_core::{
-    config_parser::Config,
+    config::Config,
+    db::DatabaseConnection,
     endpoints::{
         datasets::InsertDatapointParams,
         dynamic_evaluation_run::{
@@ -26,7 +27,7 @@ use tensorzero_core::{
         validate_tags,
     },
     error::{Error, ErrorDetails},
-    gateway_util::{setup_clickhouse, setup_http_client, AppStateData},
+    gateway_util::{setup_clickhouse, setup_http_client, GatewayHandle},
 };
 use thiserror::Error;
 use tokio::{sync::Mutex, time::error::Elapsed};
@@ -45,7 +46,7 @@ pub use client_inference_params::{ClientInferenceParams, ClientSecretString};
 pub use client_input::{ClientInput, ClientInputMessage, ClientInputMessageContent};
 
 pub use tensorzero_core::cache::CacheParamsOptions;
-pub use tensorzero_core::clickhouse::query_builder::{
+pub use tensorzero_core::db::clickhouse::query_builder::{
     BooleanMetricNode, FloatComparisonOperator, FloatMetricNode, InferenceFilterTreeNode,
     InferenceOutputSource, ListInferencesParams, TagComparisonOperator, TagNode,
     TimeComparisonOperator, TimeNode,
@@ -122,7 +123,7 @@ impl HTTPGateway {
 }
 
 struct EmbeddedGateway {
-    state: AppStateData,
+    handle: GatewayHandle,
 }
 
 /// Used to construct a `Client`
@@ -179,7 +180,12 @@ pub enum ClientBuilderError {
     #[error("Failed to configure ClickHouse: {0}")]
     Clickhouse(TensorZeroError),
     #[error("Failed to parse config: {0}")]
-    ConfigParsing(TensorZeroError),
+    ConfigParsingPreGlob(TensorZeroError),
+    #[error("Failed to parse config: {error}. Config file glob `{glob}` resolved to the following files:\n{paths}", glob = glob.glob,paths = glob.paths.iter().map(|p| p.display().to_string()).collect::<Vec<_>>().join("\n"))]
+    ConfigParsing {
+        error: TensorZeroError,
+        glob: ConfigFileGlob,
+    },
     #[error("Failed to build HTTP client: {0}")]
     HTTPClientBuild(TensorZeroError),
     #[error("Failed to get gateway version: {0}")]
@@ -215,6 +221,10 @@ pub enum ClientBuilderMode {
         /// If this timeout is hit, any in-progress LLM requests may be aborted.
         timeout: Option<std::time::Duration>,
         verify_credentials: bool,
+        // Allow turning on batch writes - used in e2e tests.
+        // We don't expose this through the Python client, since we're having deadlock issues
+        // there.
+        allow_batch_writes: bool,
     },
 }
 
@@ -252,7 +262,7 @@ impl ClientBuilder {
         match &self.mode {
             ClientBuilderMode::HTTPGateway { .. } => {
                 let client = self.build_http()?;
-                if let ClientMode::HTTPGateway(mode) = &client.mode {
+                if let ClientMode::HTTPGateway(mode) = &*client.mode {
                     mode.discover_initialize_gateway_version(&client).await?;
                 }
                 Ok(client)
@@ -262,24 +272,47 @@ impl ClientBuilder {
                 clickhouse_url,
                 timeout,
                 verify_credentials,
+                allow_batch_writes,
             } => {
                 let config = if let Some(config_file) = config_file {
+                    let glob = ConfigFileGlob::new(config_file.to_string_lossy().to_string())
+                        .map_err(|e| {
+                            ClientBuilderError::ConfigParsingPreGlob(TensorZeroError::Other {
+                                source: e.into(),
+                            })
+                        })?;
                     Arc::new(
                         Config::load_from_path_optional_verify_credentials(
-                            config_file,
+                            &glob,
                             *verify_credentials,
                         )
                         .await
                         .map_err(|e| {
-                            ClientBuilderError::ConfigParsing(TensorZeroError::Other {
-                                source: e.into(),
-                            })
+                            ClientBuilderError::ConfigParsing {
+                                error: TensorZeroError::Other { source: e.into() },
+                                glob,
+                            }
                         })?,
                     )
                 } else {
                     tracing::info!("No config file provided, so only default functions will be available. Set `config_file` to specify your `tensorzero.toml`");
                     Arc::new(Config::default())
                 };
+                if !allow_batch_writes
+                    && config.gateway.observability.batch_writes.enabled
+                    && !config
+                        .gateway
+                        .observability
+                        .batch_writes
+                        .__force_allow_embedded_batch_writes
+                {
+                    return Err(ClientBuilderError::Clickhouse(TensorZeroError::Other {
+                        source: tensorzero_core::error::Error::new(ErrorDetails::Config {
+                            message: "[gateway.observability.batch_writes] is not yet supported in embedded gateway mode".to_string(),
+                        })
+                        .into(),
+                    }));
+                }
                 let clickhouse_connection_info =
                     setup_clickhouse(&config, clickhouse_url.clone(), true)
                         .await
@@ -288,7 +321,7 @@ impl ClientBuilder {
                                 source: e.into(),
                             })
                         })?;
-                setup_howdy(clickhouse_connection_info.clone());
+
                 let http_client = if let Some(http_client) = self.http_client {
                     http_client
                 } else {
@@ -299,16 +332,16 @@ impl ClientBuilder {
                     })?
                 };
                 Ok(Client {
-                    mode: ClientMode::EmbeddedGateway {
+                    mode: Arc::new(ClientMode::EmbeddedGateway {
                         gateway: EmbeddedGateway {
-                            state: AppStateData::new_with_clickhouse_and_http_client(
+                            handle: GatewayHandle::new_with_clickhouse_and_http_client(
                                 config,
                                 clickhouse_connection_info,
                                 http_client,
                             ),
                         },
                         timeout: *timeout,
-                    },
+                    }),
                     verbose_errors: self.verbose_errors,
                     #[cfg(feature = "e2e_tests")]
                     last_body: Default::default(),
@@ -317,13 +350,36 @@ impl ClientBuilder {
         }
     }
 
-    #[cfg(any(test, feature = "e2e_tests"))]
-    pub async fn build_from_state(state: AppStateData) -> Result<Client, ClientBuilderError> {
-        Ok(Client {
-            mode: ClientMode::EmbeddedGateway {
-                gateway: EmbeddedGateway { state },
+    /// Builds a dummy client for use in pyo3. Should not otherwise be used
+    /// This avoids logging any messages
+    #[cfg(feature = "pyo3")]
+    pub fn build_dummy() -> Client {
+        use tensorzero_core::db::clickhouse::ClickHouseConnectionInfo;
+
+        Client {
+            mode: Arc::new(ClientMode::EmbeddedGateway {
+                gateway: EmbeddedGateway {
+                    handle: GatewayHandle::new_with_clickhouse_and_http_client(
+                        Arc::new(Config::default()),
+                        ClickHouseConnectionInfo::Disabled,
+                        reqwest::Client::new(),
+                    ),
+                },
                 timeout: None,
-            },
+            }),
+            verbose_errors: false,
+            #[cfg(feature = "e2e_tests")]
+            last_body: Default::default(),
+        }
+    }
+
+    #[cfg(any(test, feature = "e2e_tests"))]
+    pub async fn build_from_state(handle: GatewayHandle) -> Result<Client, ClientBuilderError> {
+        Ok(Client {
+            mode: Arc::new(ClientMode::EmbeddedGateway {
+                gateway: EmbeddedGateway { handle },
+                timeout: None,
+            }),
             verbose_errors: false,
             #[cfg(feature = "e2e_tests")]
             last_body: Default::default(),
@@ -343,11 +399,11 @@ impl ClientBuilder {
             url.set_path(&format!("{}/", url.path()));
         }
         Ok(Client {
-            mode: ClientMode::HTTPGateway(HTTPGateway {
+            mode: Arc::new(ClientMode::HTTPGateway(HTTPGateway {
                 base_url: url,
                 http_client: self.http_client.unwrap_or_default(),
                 gateway_version: Mutex::new(None),
-            }),
+            })),
             verbose_errors: self.verbose_errors,
             #[cfg(feature = "e2e_tests")]
             last_body: Default::default(),
@@ -358,7 +414,7 @@ impl ClientBuilder {
 /// A TensorZero client. This is constructed using `ClientBuilder`
 #[derive(Debug)]
 pub struct Client {
-    mode: ClientMode,
+    mode: Arc<ClientMode>,
     verbose_errors: bool,
     #[cfg(feature = "e2e_tests")]
     pub last_body: Mutex<Option<String>>,
@@ -368,13 +424,14 @@ impl Client {
     /// Queries the health of the ClickHouse database
     /// This does nothing in `ClientMode::HTTPGateway`
     pub async fn clickhouse_health(&self) -> Result<(), TensorZeroError> {
-        match &self.mode {
+        match &*self.mode {
             ClientMode::HTTPGateway(_) => Ok(()),
             ClientMode::EmbeddedGateway {
                 gateway,
                 timeout: _,
             } => gateway
-                .state
+                .handle
+                .app_state
                 .clickhouse_connection_info
                 .health()
                 .await
@@ -388,7 +445,7 @@ impl Client {
         &self,
         params: FeedbackParams,
     ) -> Result<FeedbackResponse, TensorZeroError> {
-        match &self.mode {
+        match &*self.mode {
             ClientMode::HTTPGateway(client) => {
                 let url = client
                     .base_url
@@ -406,12 +463,14 @@ impl Client {
             }
             ClientMode::EmbeddedGateway { gateway, timeout } => {
                 Ok(with_embedded_timeout(*timeout, async {
-                    tensorzero_core::endpoints::feedback::feedback(gateway.state.clone(), params)
-                        .await
-                        .map_err(err_to_http)
+                    tensorzero_core::endpoints::feedback::feedback(
+                        gateway.handle.app_state.clone(),
+                        params,
+                    )
+                    .await
+                    .map_err(err_to_http)
                 })
-                .await?
-                .0)
+                .await?)
             }
         }
     }
@@ -422,7 +481,7 @@ impl Client {
         &self,
         mut params: ClientInferenceParams,
     ) -> Result<InferenceOutput, TensorZeroError> {
-        match &self.mode {
+        match &*self.mode {
             ClientMode::HTTPGateway(client) => {
                 let gateway_version = { client.gateway_version.lock().await.clone() };
                 // We only perform this adjustment in HTTP gateway mode, since the embedded gateway
@@ -481,15 +540,28 @@ impl Client {
                 }
             }
             ClientMode::EmbeddedGateway { gateway, timeout } => {
+                let client = self.mode.clone();
                 Ok(with_embedded_timeout(*timeout, async {
-                    tensorzero_core::endpoints::inference::inference(
-                        gateway.state.config.clone(),
-                        &gateway.state.http_client,
-                        gateway.state.clickhouse_connection_info.clone(),
+                    let res = tensorzero_core::endpoints::inference::inference(
+                        gateway.handle.app_state.config.clone(),
+                        &gateway.handle.app_state.http_client,
+                        gateway.handle.app_state.clickhouse_connection_info.clone(),
                         params.try_into().map_err(err_to_http)?,
+                        // Make the stream hold on to a reference to the client,
+                        // so that we client is only dropped after all streams have finished
+                        // See 'create_stream' and 'GatewayHandle' for more details
+                        client,
                     )
                     .await
-                    .map_err(err_to_http)
+                    .map_err(err_to_http)?;
+                    match res {
+                        InferenceOutput::NonStreaming(response) => {
+                            Ok(InferenceOutput::NonStreaming(response))
+                        }
+                        InferenceOutput::Streaming(stream) => {
+                            Ok(InferenceOutput::Streaming(stream))
+                        }
+                    }
                 })
                 .await?)
             }
@@ -500,7 +572,7 @@ impl Client {
         &self,
         storage_path: StoragePath,
     ) -> Result<ObjectResponse, TensorZeroError> {
-        match &self.mode {
+        match &*self.mode {
             ClientMode::HTTPGateway(client) => {
                 let url = client
                     .base_url
@@ -531,7 +603,7 @@ impl Client {
             ClientMode::EmbeddedGateway { gateway, timeout } => {
                 Ok(with_embedded_timeout(*timeout, async {
                     tensorzero_core::endpoints::object_storage::get_object(
-                        &gateway.state.config,
+                        &gateway.handle.app_state.config,
                         storage_path,
                     )
                     .await
@@ -555,7 +627,7 @@ impl Client {
         }
         // Set internal to true so we don't validate the tags again
         params.internal = true;
-        match &self.mode {
+        match &*self.mode {
             ClientMode::HTTPGateway(client) => {
                 let url = client.base_url.join("dynamic_evaluation_run").map_err(|e| TensorZeroError::Other {
                     source: tensorzero_core::error::Error::new(ErrorDetails::InvalidBaseUrl {
@@ -569,7 +641,7 @@ impl Client {
             ClientMode::EmbeddedGateway { gateway, timeout } => {
                 Ok(with_embedded_timeout(*timeout, async {
                     tensorzero_core::endpoints::dynamic_evaluation_run::dynamic_evaluation_run(
-                        gateway.state.clone(),
+                        gateway.handle.app_state.clone(),
                         params,
                     )
                     .await
@@ -585,7 +657,7 @@ impl Client {
         run_id: Uuid,
         params: DynamicEvaluationRunEpisodeParams,
     ) -> Result<DynamicEvaluationRunEpisodeResponse, TensorZeroError> {
-        match &self.mode {
+        match &*self.mode {
             ClientMode::HTTPGateway(client) => {
                 let url = client.base_url.join(&format!("dynamic_evaluation_run/{run_id}/episode")).map_err(|e| TensorZeroError::Other {
                     source: tensorzero_core::error::Error::new(ErrorDetails::InvalidBaseUrl {
@@ -599,7 +671,7 @@ impl Client {
             ClientMode::EmbeddedGateway { gateway, timeout } => {
                 Ok(with_embedded_timeout(*timeout, async {
                     tensorzero_core::endpoints::dynamic_evaluation_run::dynamic_evaluation_run_episode(
-                        gateway.state.clone(),
+                        gateway.handle.app_state.clone(),
                         run_id,
                         params,
                     )
@@ -616,7 +688,7 @@ impl Client {
         dataset_name: String,
         params: InsertDatapointParams,
     ) -> Result<Vec<Uuid>, TensorZeroError> {
-        match &self.mode {
+        match &*self.mode {
             ClientMode::HTTPGateway(client) => {
                 let url = client.base_url.join(&format!("datasets/{dataset_name}/datapoints/bulk")).map_err(|e| TensorZeroError::Other {
                     source: tensorzero_core::error::Error::new(ErrorDetails::InvalidBaseUrl {
@@ -632,9 +704,9 @@ impl Client {
                     tensorzero_core::endpoints::datasets::insert_datapoint(
                         dataset_name,
                         params,
-                        &gateway.state.config,
-                        &gateway.state.http_client,
-                        &gateway.state.clickhouse_connection_info,
+                        &gateway.handle.app_state.config,
+                        &gateway.handle.app_state.http_client,
+                        &gateway.handle.app_state.clickhouse_connection_info,
                     )
                     .await
                     .map_err(err_to_http)
@@ -649,7 +721,7 @@ impl Client {
         dataset_name: String,
         datapoint_id: Uuid,
     ) -> Result<(), TensorZeroError> {
-        match &self.mode {
+        match &*self.mode {
             ClientMode::HTTPGateway(client) => {
                 let url = client.base_url.join(&format!("datasets/{dataset_name}/datapoints/{datapoint_id}")).map_err(|e| TensorZeroError::Other {
                     source: tensorzero_core::error::Error::new(ErrorDetails::InvalidBaseUrl {
@@ -683,7 +755,7 @@ impl Client {
                     tensorzero_core::endpoints::datasets::delete_datapoint(
                         dataset_name,
                         datapoint_id,
-                        &gateway.state.clickhouse_connection_info,
+                        &gateway.handle.app_state.clickhouse_connection_info,
                     )
                     .await
                     .map_err(err_to_http)
@@ -700,7 +772,7 @@ impl Client {
         limit: Option<u32>,
         offset: Option<u32>,
     ) -> Result<Vec<Datapoint>, TensorZeroError> {
-        match &self.mode {
+        match &*self.mode {
             ClientMode::HTTPGateway(client) => {
                 let url = client.base_url.join(&format!("datasets/{dataset_name}/datapoints")).map_err(|e| TensorZeroError::Other {
                     source: tensorzero_core::error::Error::new(ErrorDetails::InvalidBaseUrl {
@@ -721,7 +793,7 @@ impl Client {
                 Ok(with_embedded_timeout(*timeout, async {
                     tensorzero_core::endpoints::datasets::list_datapoints(
                         dataset_name,
-                        &gateway.state.clickhouse_connection_info,
+                        &gateway.handle.app_state.clickhouse_connection_info,
                         function_name,
                         limit,
                         offset,
@@ -739,7 +811,7 @@ impl Client {
         dataset_name: String,
         datapoint_id: Uuid,
     ) -> Result<Datapoint, TensorZeroError> {
-        match &self.mode {
+        match &*self.mode {
             ClientMode::HTTPGateway(client) => {
                 let url = client.base_url.join(&format!("datasets/{dataset_name}/datapoints/{datapoint_id}")).map_err(|e| TensorZeroError::Other {
                     source: tensorzero_core::error::Error::new(ErrorDetails::InvalidBaseUrl {
@@ -755,7 +827,7 @@ impl Client {
                     tensorzero_core::endpoints::datasets::get_datapoint(
                         dataset_name,
                         datapoint_id,
-                        &gateway.state.clickhouse_connection_info,
+                        &gateway.handle.app_state.clickhouse_connection_info,
                     )
                     .await
                     .map_err(err_to_http)
@@ -772,11 +844,11 @@ impl Client {
         &self,
         dataset_name: String,
     ) -> Result<StaleDatasetResponse, TensorZeroError> {
-        match &self.mode {
+        match &*self.mode {
             ClientMode::EmbeddedGateway { gateway, timeout } => {
                 with_embedded_timeout(*timeout, async {
                     tensorzero_core::endpoints::datasets::stale_dataset(
-                        &gateway.state.clickhouse_connection_info,
+                        &gateway.handle.app_state.clickhouse_connection_info,
                         &dataset_name,
                     )
                     .await
@@ -815,7 +887,7 @@ impl Client {
         params: ListInferencesParams<'_>,
     ) -> Result<Vec<StoredInference>, TensorZeroError> {
         // TODO: consider adding a flag that returns the generated sql query
-        let ClientMode::EmbeddedGateway { gateway, .. } = &self.mode else {
+        let ClientMode::EmbeddedGateway { gateway, .. } = &*self.mode else {
             return Err(TensorZeroError::Other {
                 source: tensorzero_core::error::Error::new(ErrorDetails::InvalidClientMode {
                     mode: "Http".to_string(),
@@ -825,9 +897,10 @@ impl Client {
             });
         };
         let inferences = gateway
-            .state
+            .handle
+            .app_state
             .clickhouse_connection_info
-            .list_inferences(&gateway.state.config, &params)
+            .list_inferences(&gateway.handle.app_state.config, &params)
             .await
             .map_err(err_to_http)?;
         Ok(inferences)
@@ -846,7 +919,7 @@ impl Client {
         stored_samples: Vec<T>,
         variants: HashMap<String, String>, // Map from function name to variant name
     ) -> Result<Vec<RenderedSample>, TensorZeroError> {
-        let ClientMode::EmbeddedGateway { gateway, .. } = &self.mode else {
+        let ClientMode::EmbeddedGateway { gateway, .. } = &*self.mode else {
             return Err(TensorZeroError::Other {
                 source: tensorzero_core::error::Error::new(ErrorDetails::InvalidClientMode {
                     mode: "Http".to_string(),
@@ -855,9 +928,13 @@ impl Client {
                 .into(),
             });
         };
-        render_samples(gateway.state.config.clone(), stored_samples, variants)
-            .await
-            .map_err(err_to_http)
+        render_samples(
+            gateway.handle.app_state.config.clone(),
+            stored_samples,
+            variants,
+        )
+        .await
+        .map_err(err_to_http)
     }
 
     /// Launch an optimization job.
@@ -865,11 +942,11 @@ impl Client {
         &self,
         params: tensorzero_core::endpoints::optimization::LaunchOptimizationParams,
     ) -> Result<OptimizationJobHandle, TensorZeroError> {
-        match &self.mode {
+        match &*self.mode {
             ClientMode::EmbeddedGateway { gateway, timeout } => {
                 // TODO: do we want this?
                 Ok(with_embedded_timeout(*timeout, async {
-                    launch_optimization(&gateway.state.http_client, params)
+                    launch_optimization(&gateway.handle.app_state.http_client, params)
                         .await
                         .map_err(err_to_http)
                 })
@@ -891,13 +968,13 @@ impl Client {
         &self,
         params: LaunchOptimizationWorkflowParams,
     ) -> Result<OptimizationJobHandle, TensorZeroError> {
-        match &self.mode {
+        match &*self.mode {
             ClientMode::EmbeddedGateway { gateway, timeout } => {
                 with_embedded_timeout(*timeout, async {
                     launch_optimization_workflow(
-                        &gateway.state.http_client,
-                        gateway.state.config.clone(),
-                        &gateway.state.clickhouse_connection_info,
+                        &gateway.handle.app_state.http_client,
+                        gateway.handle.app_state.config.clone(),
+                        &gateway.handle.app_state.clickhouse_connection_info,
                         params,
                     )
                     .await
@@ -943,11 +1020,11 @@ impl Client {
         &self,
         job_handle: &OptimizationJobHandle,
     ) -> Result<OptimizationJobInfo, TensorZeroError> {
-        match &self.mode {
+        match &*self.mode {
             ClientMode::EmbeddedGateway { gateway, timeout } => {
                 Ok(with_embedded_timeout(*timeout, async {
                     tensorzero_core::endpoints::optimization::poll_optimization(
-                        &gateway.state.http_client,
+                        &gateway.handle.app_state.http_client,
                         job_handle,
                     )
                     .await
@@ -977,8 +1054,10 @@ impl Client {
     }
 
     pub fn get_config(&self) -> Result<Arc<Config>, TensorZeroError> {
-        match &self.mode {
-            ClientMode::EmbeddedGateway { gateway, .. } => Ok(gateway.state.config.clone()),
+        match &*self.mode {
+            ClientMode::EmbeddedGateway { gateway, .. } => {
+                Ok(gateway.handle.app_state.config.clone())
+            }
             ClientMode::HTTPGateway(_) => Err(TensorZeroError::Other {
                 source: tensorzero_core::error::Error::new(ErrorDetails::InvalidClientMode {
                     mode: "Http".to_string(),
@@ -1153,9 +1232,9 @@ impl Client {
     }
 
     #[cfg(any(feature = "e2e_tests", feature = "pyo3"))]
-    pub fn get_app_state_data(&self) -> Option<&AppStateData> {
-        match &self.mode {
-            ClientMode::EmbeddedGateway { gateway, .. } => Some(&gateway.state),
+    pub fn get_app_state_data(&self) -> Option<&tensorzero_core::gateway_util::AppStateData> {
+        match &*self.mode {
+            ClientMode::EmbeddedGateway { gateway, .. } => Some(&gateway.handle.app_state),
             _ => None,
         }
     }
@@ -1181,11 +1260,11 @@ impl Client {
     }
 
     async fn update_gateway_version(&self, version: String) {
-        match &self.mode {
+        match &*self.mode {
             ClientMode::HTTPGateway(client) => {
                 // Acquire the lock on the gateway version
                 let mut gateway_version = client.gateway_version.lock().await;
-                *gateway_version = Some(version)
+                *gateway_version = Some(version);
             }
             // Should never be called
             ClientMode::EmbeddedGateway { .. } => {}
@@ -1194,7 +1273,7 @@ impl Client {
 
     #[cfg(feature = "e2e_tests")]
     pub async fn get_gateway_version(&self) -> Option<String> {
-        match &self.mode {
+        match &*self.mode {
             ClientMode::HTTPGateway(client) => client.gateway_version.lock().await.clone(),
             ClientMode::EmbeddedGateway { .. } => None,
         }
@@ -1219,9 +1298,13 @@ async fn with_embedded_timeout<R, F: Future<Output = Result<R, TensorZeroError>>
 /// and returns a `TensorZeroError` instead of a `ConfigError`.
 /// This function does NOT verify credentials.
 pub async fn get_config_no_verify_credentials(path: PathBuf) -> Result<Config, TensorZeroError> {
-    Config::load_from_path_optional_verify_credentials(&path, false)
-        .await
-        .map_err(|e| TensorZeroError::Other { source: e.into() })
+    Config::load_from_path_optional_verify_credentials(
+        &ConfigFileGlob::new(path.to_string_lossy().to_string())
+            .map_err(|e| TensorZeroError::Other { source: e.into() })?,
+        false,
+    )
+    .await
+    .map_err(|e| TensorZeroError::Other { source: e.into() })
 }
 
 /// Compares two TensorZero version strings, returning `None`
@@ -1314,23 +1397,21 @@ mod tests {
     use tracing_test::traced_test;
 
     #[tokio::test]
-    #[ignore] // TODO - set an environment variable, or create a new config with dummy credentials
     async fn test_missing_clickhouse() {
         // This config file requires ClickHouse, so it should fail if no ClickHouse URL is provided
         let err = ClientBuilder::new(ClientBuilderMode::EmbeddedGateway {
-            config_file: Some(PathBuf::from(
-                "../../examples/haiku-hidden-preferences/config/tensorzero.toml",
-            )),
+            config_file: Some(PathBuf::from("tests/test_config.toml")),
             clickhouse_url: None,
             timeout: None,
             verify_credentials: true,
+            allow_batch_writes: true,
         })
         .build()
         .await
         .expect_err("ClientBuilder should have failed");
         let err_msg = err.to_string();
         assert!(
-            err.to_string().contains("Missing ClickHouse URL"),
+            err_msg.contains("Missing environment variable TENSORZERO_CLICKHOUSE_URL"),
             "Bad error message: {err_msg}"
         );
     }
@@ -1346,6 +1427,7 @@ mod tests {
             clickhouse_url: None,
             timeout: None,
             verify_credentials: true,
+            allow_batch_writes: true,
         })
         .build()
         .await
@@ -1364,6 +1446,7 @@ mod tests {
             clickhouse_url: None,
             timeout: None,
             verify_credentials: true,
+            allow_batch_writes: true,
         })
         .build()
         .await
