@@ -1,9 +1,9 @@
 use std::collections::HashMap;
 use std::sync::Arc;
 
-use crate::clickhouse::ClickHouseConnectionInfo;
-use crate::embeddings::{EmbeddingRequest, EmbeddingResponse};
-use crate::error::{Error, ErrorDetails};
+use crate::db::clickhouse::{ClickHouseConnectionInfo, TableName};
+use crate::embeddings::{EmbeddingModelResponse, EmbeddingRequest};
+use crate::error::{warn_discarded_cache_write, Error, ErrorDetails};
 use crate::inference::types::file::serialize_with_file_data;
 use crate::inference::types::{
     ContentBlockChunk, ContentBlockOutput, FinishReason, ModelInferenceRequest,
@@ -13,11 +13,14 @@ use crate::model::StreamResponse;
 use crate::serde_util::deserialize_json_string;
 use blake3::Hash;
 use clap::ValueEnum;
-use serde::de::DeserializeOwned;
+use serde::de::{DeserializeOwned, IgnoredAny};
 use serde::{Deserialize, Serialize};
 use std::fmt::Debug;
 
-#[derive(Clone, Copy, Debug, Default, Deserialize, Eq, PartialEq, Serialize, ValueEnum)]
+#[derive(
+    Clone, Copy, Debug, Default, Deserialize, Eq, PartialEq, Serialize, ValueEnum, ts_rs::TS,
+)]
+#[ts(export)]
 #[serde(rename_all = "snake_case")]
 #[clap(rename_all = "snake_case")]
 pub enum CacheEnabledMode {
@@ -38,7 +41,8 @@ impl CacheEnabledMode {
     }
 }
 
-#[derive(Clone, Debug, Default, Deserialize, PartialEq, Serialize)]
+#[derive(Clone, Debug, Default, Deserialize, PartialEq, Serialize, ts_rs::TS)]
+#[ts(export)]
 pub struct CacheParamsOptions {
     #[serde(default)]
     pub max_age_s: Option<u32>,
@@ -210,11 +214,39 @@ pub struct CacheData<T: CacheOutput> {
 /// to/from ClickHouse
 /// We use a marker trait rather than an enum so that the expected type can be enforced by the caller
 /// (e.g. `infer_stream` will never try to deserialize a `NonStreamingCacheData`)
-pub trait CacheOutput {}
+pub trait CacheOutput {
+    /// If this return `false`, then we'll log a warning and skip writing this entry to the cache
+    fn should_write_to_cache(&self) -> bool;
+}
 
-impl CacheOutput for StreamingCacheData {}
-impl CacheOutput for NonStreamingCacheData {}
-impl CacheOutput for EmbeddingCacheData {}
+impl CacheOutput for StreamingCacheData {
+    fn should_write_to_cache(&self) -> bool {
+        true
+    }
+}
+impl CacheOutput for NonStreamingCacheData {
+    fn should_write_to_cache(&self) -> bool {
+        for block in &self.blocks {
+            if let ContentBlockOutput::ToolCall(tool_call) = block {
+                // We skip writing to the cache if the tool call arguments are not valid JSON
+                // We're assuming that it's almost never useful to have an invalid tool call cached
+                // (in particular, tensorzero is not being used with a provider/model that only ever
+                // emits invalid json for its tool call arguments).
+                // The invalid tool call will still be returned to the user, but we won't create a
+                // cache entry, even if the user turned on caching.
+                if serde_json::from_str::<IgnoredAny>(&tool_call.arguments).is_err() {
+                    return false;
+                }
+            }
+        }
+        true
+    }
+}
+impl CacheOutput for EmbeddingCacheData {
+    fn should_write_to_cache(&self) -> bool {
+        true
+    }
+}
 
 #[derive(Debug, Deserialize, Serialize)]
 #[serde(transparent)]
@@ -237,6 +269,24 @@ pub struct StreamingCacheData {
     pub chunks: Vec<CachedProviderInferenceResponseChunk>,
 }
 
+fn spawn_maybe_cache_write<T: Serialize + CacheOutput + Send + Sync + 'static>(
+    row: FullCacheRow<T>,
+    clickhouse_client: ClickHouseConnectionInfo,
+) {
+    tokio::spawn(async move {
+        if row.data.output.should_write_to_cache() {
+            if let Err(e) = clickhouse_client
+                .write_batched(&[row], TableName::ModelInferenceCache)
+                .await
+            {
+                tracing::warn!("Failed to write to cache: {e}");
+            }
+        } else {
+            warn_discarded_cache_write(&row.data.raw_response);
+        }
+    });
+}
+
 // This doesn't block
 pub fn start_cache_write<T: Serialize + CacheOutput + Send + Sync + 'static>(
     clickhouse_client: &ClickHouseConnectionInfo,
@@ -255,28 +305,21 @@ pub fn start_cache_write<T: Serialize + CacheOutput + Send + Sync + 'static>(
     let output_tokens = usage.output_tokens;
     let clickhouse_client = clickhouse_client.clone();
     let finish_reason = finish_reason.cloned();
-    tokio::spawn(async move {
-        if let Err(e) = clickhouse_client
-            .write(
-                &[FullCacheRow {
-                    short_cache_key,
-                    long_cache_key,
-                    data: CacheData {
-                        output,
-                        raw_request,
-                        raw_response,
-                        input_tokens,
-                        output_tokens,
-                        finish_reason,
-                    },
-                }],
-                "ModelInferenceCache",
-            )
-            .await
-        {
-            tracing::warn!("Failed to write to cache: {e}");
-        }
-    });
+    spawn_maybe_cache_write(
+        FullCacheRow {
+            short_cache_key,
+            long_cache_key,
+            data: CacheData {
+                output,
+                raw_request,
+                raw_response,
+                input_tokens,
+                output_tokens,
+                finish_reason,
+            },
+        },
+        clickhouse_client,
+    );
     Ok(())
 }
 
@@ -322,25 +365,21 @@ pub fn start_cache_write_streaming(
     };
     let raw_request = raw_request.to_string();
     let clickhouse_client = clickhouse_client.clone();
-    tokio::spawn(async move {
-        clickhouse_client
-            .write(
-                &[FullCacheRow {
-                    short_cache_key,
-                    long_cache_key,
-                    data: CacheData {
-                        output,
-                        raw_request,
-                        raw_response: String::new(),
-                        input_tokens,
-                        output_tokens,
-                        finish_reason,
-                    },
-                }],
-                "ModelInferenceCache",
-            )
-            .await
-    });
+    spawn_maybe_cache_write(
+        FullCacheRow {
+            short_cache_key,
+            long_cache_key,
+            data: CacheData {
+                output,
+                raw_request,
+                raw_response: String::new(),
+                input_tokens,
+                output_tokens,
+                finish_reason,
+            },
+        },
+        clickhouse_client,
+    );
     Ok(())
 }
 
@@ -348,14 +387,14 @@ pub async fn embedding_cache_lookup(
     clickhouse_connection_info: &ClickHouseConnectionInfo,
     request: &EmbeddingModelProviderRequest<'_>,
     max_age_s: Option<u32>,
-) -> Result<Option<EmbeddingResponse>, Error> {
+) -> Result<Option<EmbeddingModelResponse>, Error> {
     let result = cache_lookup_inner::<EmbeddingCacheData>(
         clickhouse_connection_info,
         request.get_cache_key()?,
         max_age_s,
     )
     .await?;
-    Ok(result.map(|result| EmbeddingResponse::from_cache(result, request)))
+    Ok(result.map(|result| EmbeddingModelResponse::from_cache(result, request)))
 }
 
 pub async fn cache_lookup(
@@ -399,7 +438,7 @@ pub async fn cache_lookup_inner<T: CacheOutput + DeserializeOwned>(
     let long_cache_key = cache_key.get_long_key();
     // The clickhouse query args look like rust format string args, but they're not.
     let query = if max_age_s.is_some() {
-        r#"
+        r"
             SELECT
                 output,
                 raw_request,
@@ -414,9 +453,9 @@ pub async fn cache_lookup_inner<T: CacheOutput + DeserializeOwned>(
             ORDER BY timestamp DESC
             LIMIT 1
             FORMAT JSONEachRow
-        "#
+        "
     } else {
-        r#"
+        r"
             SELECT
                 output,
                 raw_request,
@@ -430,7 +469,7 @@ pub async fn cache_lookup_inner<T: CacheOutput + DeserializeOwned>(
             ORDER BY timestamp DESC
             LIMIT 1
             FORMAT JSONEachRow
-        "#
+        "
     };
     let mut query_params = HashMap::from([
         ("short_cache_key", short_cache_key.as_str()),
