@@ -9,7 +9,8 @@ use object_store::{ObjectStore, PutMode, PutOptions};
 use secrecy::SecretString;
 use serde::{Deserialize, Serialize};
 use serde_json::Value;
-use std::collections::HashMap;
+use std::borrow::Cow;
+use std::collections::{BTreeMap, HashMap};
 use std::future::Future;
 use std::pin::Pin;
 use std::sync::Arc;
@@ -17,11 +18,12 @@ use std::time::Duration;
 use tokio::time::Instant;
 use tokio_stream::StreamExt;
 use tracing::instrument;
+use tracing_opentelemetry::OpenTelemetrySpanExt;
 use uuid::Uuid;
 
 use crate::cache::{CacheOptions, CacheParamsOptions};
-use crate::clickhouse::ClickHouseConnectionInfo;
-use crate::config_parser::{Config, ObjectStoreInfo};
+use crate::config::{Config, ObjectStoreInfo, SchemaData, UninitializedVariantInfo};
+use crate::db::clickhouse::{ClickHouseConnectionInfo, TableName};
 use crate::embeddings::EmbeddingModelTable;
 use crate::error::{Error, ErrorDetails};
 use crate::function::FunctionConfig;
@@ -40,9 +42,11 @@ use crate::inference::types::{
     ResolvedInputMessageContent, Usage,
 };
 use crate::jsonschema_util::DynamicJSONSchema;
+use crate::minijinja_util::TemplateConfig;
 use crate::model::ModelTable;
 use crate::tool::{DynamicToolParams, ToolCallConfig, ToolChoice};
 use crate::variant::chat_completion::ChatCompletionConfig;
+use crate::variant::dynamic::load_dynamic_variant_info;
 use crate::variant::{InferenceConfig, JsonMode, Variant, VariantConfig, VariantInfo};
 
 use super::dynamic_evaluation_run::validate_inference_episode_id_and_apply_dynamic_evaluation_run;
@@ -106,6 +110,8 @@ pub struct Params {
     pub extra_body: UnfilteredInferenceExtraBody,
     #[serde(default)]
     pub extra_headers: UnfilteredInferenceExtraHeaders,
+    #[serde(default)]
+    pub internal_dynamic_variant_config: Option<UninitializedVariantInfo>,
 }
 
 #[derive(Clone, Debug)]
@@ -148,7 +154,7 @@ pub async fn inference_handler(
     StructuredJson(params): StructuredJson<Params>,
 ) -> Result<Response<Body>, Error> {
     let inference_output =
-        inference(config, &http_client, clickhouse_connection_info, params).await?;
+        inference(config, &http_client, clickhouse_connection_info, params, ()).await?;
     match inference_output {
         InferenceOutput::NonStreaming(response) => Ok(Json(response).into_response()),
         InferenceOutput::Streaming(stream) => {
@@ -188,7 +194,7 @@ pub struct InferenceIds {
 
 #[instrument(
     name="inference",
-    skip(config, http_client, clickhouse_connection_info, params),
+    skip(config, http_client, clickhouse_connection_info, params, extra_handle),
     fields(
         function_name,
         model_name,
@@ -198,11 +204,13 @@ pub struct InferenceIds {
         otel.name = "function_inference"
     )
 )]
-pub async fn inference(
+pub async fn inference<T: Send + 'static>(
     config: Arc<Config>,
     http_client: &reqwest::Client,
     clickhouse_connection_info: ClickHouseConnectionInfo,
     params: Params,
+    // See 'create_stream' for more details about this parameter
+    extra_handle: T,
 ) -> Result<InferenceOutput, Error> {
     let span = tracing::Span::current();
     if let Some(function_name) = &params.function_name {
@@ -217,6 +225,9 @@ pub async fn inference(
     if let Some(episode_id) = &params.episode_id {
         span.record("episode_id", episode_id.to_string());
     }
+    for (tag_key, tag_value) in &params.tags {
+        span.set_attribute(format!("tags.{tag_key}"), tag_value.clone());
+    }
     // To be used for the Inference table processing_time measurements
     let start_time = Instant::now();
     let inference_id = Uuid::now_v7();
@@ -234,15 +245,17 @@ pub async fn inference(
         &clickhouse_connection_info,
     )
     .await?;
-    tracing::Span::current().record("episode_id", episode_id.to_string());
+    // Record the episode id if we didn't already have one
+    if params.episode_id.is_none() {
+        tracing::Span::current().record("episode_id", episode_id.to_string());
+    }
 
     let (function, function_name) = find_function(&params, &config)?;
-    // Collect the function variant names as a Vec<&str>
-    let mut candidate_variant_names: Vec<&str> =
-        function.variants().keys().map(AsRef::as_ref).collect();
+    let mut candidate_variants: BTreeMap<String, Arc<VariantInfo>> =
+        function.variants().clone().into_iter().collect();
 
     // If the function has no variants, return an error
-    if candidate_variant_names.is_empty() {
+    if candidate_variants.is_empty() {
         return Err(ErrorDetails::InvalidFunctionVariants {
             message: format!("Function `{function_name}` has no variants"),
         }
@@ -252,38 +265,30 @@ pub async fn inference(
     // Validate the input
     function.validate_inference_params(&params)?;
 
-    let tool_config = function.prepare_tool_config(params.dynamic_tool_params, &config.tools)?;
-
-    // If a variant is pinned, only that variant should be attempted
-    if let Some(ref variant_name) = params.variant_name {
-        candidate_variant_names.retain(|k| k == variant_name);
-
-        // If the pinned variant doesn't exist, return an error
-        if candidate_variant_names.is_empty() {
-            return Err(ErrorDetails::UnknownVariant {
-                name: variant_name.to_string(),
-            }
-            .into());
-        }
-        params.tags.insert(
-            "tensorzero::variant_pinned".to_string(),
-            variant_name.to_string(),
-        );
-    } else {
-        // Remove all zero-weight variants - these can only be used if explicitly pinned above
-        candidate_variant_names.retain(|name| {
-            if let Some(variant) = function.variants().get(*name) {
-                // Retain 'None' and positive-weight variants, discarding zero-weight variants
-                variant.inner.weight().is_none_or(|w| w > 0.0)
-            } else {
-                // Keep missing variants - later code will error if we try to use them
-                true
-            }
-        });
-    }
-
     // Should we store the results?
     let dryrun = params.dryrun.unwrap_or(false);
+    if params.internal_dynamic_variant_config.is_some() && !dryrun {
+        return Err(ErrorDetails::InvalidRequest {
+            message:
+                "If `internal_dynamic_variant_config` is used, `dryrun` must also be set to true"
+                    .to_string(),
+        }
+        .into());
+    }
+
+    let tool_config = function.prepare_tool_config(params.dynamic_tool_params, &config.tools)?;
+    let mut templates = Cow::Borrowed(&config.templates);
+
+    prepare_candidate_variants(
+        &mut candidate_variants,
+        &mut params.tags,
+        params.variant_name.as_deref(),
+        params.internal_dynamic_variant_config,
+        &mut templates,
+        &function,
+        function_name.clone(),
+    )?;
+    let templates = &*templates;
 
     // Increment the request count if we're not in dryrun mode
     if !dryrun {
@@ -302,24 +307,11 @@ pub async fn inference(
     let stream = params.stream.unwrap_or(false);
 
     // Keep track of which variants failed
-    let mut variant_errors = std::collections::HashMap::new();
+    let mut variant_errors: HashMap<String, Error> = HashMap::new();
 
     // Set up inference config
     let output_schema = params.output_schema.map(DynamicJSONSchema::new);
-    let mut inference_config = InferenceConfig {
-        function_name: &function_name,
-        variant_name: None,
-        templates: &config.templates,
-        tool_config: tool_config.as_ref(),
-        dynamic_output_schema: output_schema.as_ref(),
-        ids: InferenceIds {
-            inference_id,
-            episode_id,
-        },
-        extra_cache_key: None,
-        extra_body: Default::default(),
-        extra_headers: Default::default(),
-    };
+
     let inference_clients = InferenceClients {
         http_client,
         clickhouse_connection_info: &clickhouse_connection_info,
@@ -339,19 +331,25 @@ pub async fn inference(
         })
         .await?;
     // Keep sampling variants until one succeeds
-    while !candidate_variant_names.is_empty() {
-        let (variant_name, variant) = sample_variant(
-            &mut candidate_variant_names,
-            function.variants(),
-            &function_name,
-            &episode_id,
-        )?;
+    while !candidate_variants.is_empty() {
+        let (variant_name, variant) =
+            sample_variant(&mut candidate_variants, &function_name, &episode_id)?;
         // Will be edited by the variant as part of making the request so we must clone here
         let variant_inference_params = params.params.clone();
-
-        inference_config.variant_name = Some(variant_name);
-        inference_config.extra_body = params.extra_body.clone();
-        inference_config.extra_headers = params.extra_headers.clone();
+        let inference_config = InferenceConfig {
+            function_name: &function_name,
+            variant_name: &variant_name,
+            templates,
+            tool_config: tool_config.as_ref(),
+            dynamic_output_schema: output_schema.as_ref(),
+            ids: InferenceIds {
+                inference_id,
+                episode_id,
+            },
+            extra_cache_key: None,
+            extra_body: Cow::Borrowed(&params.extra_body),
+            extra_headers: Cow::Borrowed(&params.extra_headers),
+        };
         if stream {
             let result = variant
                 .infer_stream(
@@ -372,20 +370,18 @@ pub async fn inference(
                     tracing::warn!(
                         "functions.{function_name:?}.variants.{variant_name:?} failed during inference: {e}",
                         function_name = params.function_name,
-                        variant_name = variant_name,
+                        variant_name = inference_config.variant_name,
                     );
-                    variant_errors.insert(variant_name.to_string(), e);
+                    variant_errors.insert(inference_config.variant_name.to_string(), e);
                     continue;
                 }
             };
-
-            let extra_body = inference_config.extra_body.clone();
-            let extra_headers = inference_config.extra_headers.clone();
-
+            let extra_body = inference_config.extra_body.into_owned();
+            let extra_headers = inference_config.extra_headers.into_owned();
             // Create InferenceMetadata for a streaming inference
             let inference_metadata = InferenceMetadata {
                 function_name: function_name.to_string(),
-                variant_name: variant_name.to_string(),
+                variant_name: inference_config.variant_name.to_string(),
                 inference_id,
                 episode_id,
                 input: resolved_input.clone(),
@@ -414,6 +410,7 @@ pub async fn inference(
                 inference_metadata,
                 stream,
                 clickhouse_connection_info,
+                extra_handle,
             );
 
             return Ok(InferenceOutput::Streaming(Box::pin(stream)));
@@ -435,21 +432,21 @@ pub async fn inference(
                     tracing::warn!(
                         "functions.{function_name}.variants.{variant_name} failed during inference: {e}",
                         function_name = function_name,
-                        variant_name = variant_name,
+                        variant_name = inference_config.variant_name,
                     );
-                    variant_errors.insert(variant_name.to_string(), e);
+                    variant_errors.insert(inference_config.variant_name.to_string(), e);
                     continue;
                 }
             };
 
             if !dryrun {
                 // Spawn a thread for a trailing write to ClickHouse so that it doesn't block the response
-                let extra_body = inference_config.extra_body.clone();
-                let extra_headers = inference_config.extra_headers.clone();
                 let result_to_write = result.clone();
+                let extra_body = inference_config.extra_body.into_owned();
+                let extra_headers = inference_config.extra_headers.into_owned();
                 let write_metadata = InferenceDatabaseInsertMetadata {
                     function_name: function_name.to_string(),
-                    variant_name: variant_name.to_string(),
+                    variant_name: inference_config.variant_name.to_string(),
                     episode_id,
                     tool_config,
                     processing_time: Some(start_time.elapsed()),
@@ -487,7 +484,7 @@ pub async fn inference(
                 result.set_original_response(None);
             }
 
-            let response = InferenceResponse::new(result, episode_id, variant_name.to_string());
+            let response = InferenceResponse::new(result, episode_id, variant_name);
 
             return Ok(InferenceOutput::NonStreaming(response));
         }
@@ -505,13 +502,17 @@ pub async fn inference(
 /// If `model_name` is specified, then we use the special 'default' function
 /// Returns the function config and the function name
 fn find_function(params: &Params, config: &Config) -> Result<(Arc<FunctionConfig>, String), Error> {
-    match (&params.function_name, &params.model_name) {
+    match (
+        &params.function_name,
+        &params.model_name,
+        &params.internal_dynamic_variant_config,
+    ) {
         // Get the function config or return an error if it doesn't exist
-        (Some(function_name), None) => Ok((
+        (Some(function_name), None, _) => Ok((
             config.get_function(function_name)?.into_owned(),
             function_name.to_string(),
         )),
-        (None, Some(model_name)) => {
+        (None, Some(model_name), None) => {
             if params.variant_name.is_some() {
                 return Err(ErrorDetails::InvalidInferenceTarget {
                     message: "`variant_name` cannot be provided when using `model_name`"
@@ -540,9 +541,7 @@ fn find_function(params: &Params, config: &Config) -> Result<(Arc<FunctionConfig
                     )]
                     .into_iter()
                     .collect(),
-                    system_schema: None,
-                    user_schema: None,
-                    assistant_schema: None,
+                    schemas: SchemaData::default(),
                     tools: vec![],
                     tool_choice: ToolChoice::Auto,
                     parallel_tool_calls: None,
@@ -551,23 +550,34 @@ fn find_function(params: &Params, config: &Config) -> Result<(Arc<FunctionConfig
                 DEFAULT_FUNCTION_NAME.to_string(),
             ))
         }
-        (Some(_), Some(_)) => Err(ErrorDetails::InvalidInferenceTarget {
+        (Some(_), Some(_), None) => Err(ErrorDetails::InvalidInferenceTarget {
             message: "Only one of `function_name` or `model_name` can be provided".to_string(),
         }
         .into()),
-        (None, None) => Err(ErrorDetails::InvalidInferenceTarget {
+        (None, None, None) => Err(ErrorDetails::InvalidInferenceTarget {
             message: "Either `function_name` or `model_name` must be provided".to_string(),
         }
         .into()),
+        (_, _, Some(_)) => Err(ErrorDetails::InvalidInferenceTarget {
+            message: "If a dynamic variant config is passed, `function_name` must be specified."
+                .to_string(),
+        }
+        .into()),
     }
 }
 
-fn create_stream(
+fn create_stream<T: Send + 'static>(
     function: Arc<FunctionConfig>,
     config: Arc<Config>,
     metadata: InferenceMetadata,
     mut stream: InferenceResultStream,
     clickhouse_connection_info: ClickHouseConnectionInfo,
+    // An arbitrary 'handle' parameter, which is guaranteed to be dropped after `clickhouse_connection_info`
+    // This is used by the embedded Rust client to ensure that we only drop the last reference to the client
+    // after all outstanding streams have finished (so that we can block in `Drop` from Python without risking
+    // a deadlock due to a `ClickHouseConnectionInfo` being kept alive by a stream in Python
+    // See `GatewayHandle` for more details
+    extra_handle: T,
 ) -> impl Stream<Item = Result<InferenceResponseChunk, Error>> + Send {
     async_stream::stream! {
         let mut buffer = vec![];
@@ -657,7 +667,7 @@ fn create_stream(
             let config = config.clone();
             let async_write = config.gateway.observability.async_writes;
             let write_future = async move {
-                let templates = &config.templates;
+                let templates = Cow::Borrowed(&config.templates);
                 let collect_chunks_args = CollectChunksArgs {
                     value: buffer,
                     inference_id,
@@ -673,7 +683,7 @@ fn create_stream(
                     function_name: &function_name,
                     variant_name: &variant_name,
                     dynamic_output_schema,
-                    templates,
+                    templates: &templates,
                     tool_config: tool_config.as_ref(),
                     cached,
                     extra_body: extra_body.clone(),
@@ -710,6 +720,8 @@ fn create_stream(
                         ).await;
 
                 }
+                drop(clickhouse_connection_info);
+                drop(extra_handle);
             };
             if async_write {
                 tokio::spawn(write_future);
@@ -857,7 +869,7 @@ async fn write_inference(
         // Write the model responses to the ModelInference table
         for response in model_responses {
             let _ = clickhouse_connection_info
-                .write(&[response], "ModelInference")
+                .write_batched(&[response], TableName::ModelInference)
                 .await;
         }
         // Write the inference to the Inference table
@@ -866,14 +878,14 @@ async fn write_inference(
                 let chat_inference =
                     ChatInferenceDatabaseInsert::new(result, input.clone(), metadata);
                 let _ = clickhouse_connection_info
-                    .write(&[chat_inference], "ChatInference")
+                    .write_batched(&[chat_inference], TableName::ChatInference)
                     .await;
             }
             InferenceResult::Json(result) => {
                 let json_inference =
                     JsonInferenceDatabaseInsert::new(result, input.clone(), metadata);
                 let _ = clickhouse_connection_info
-                    .write(&[json_inference], "JsonInference")
+                    .write_batched(&[json_inference], TableName::JsonInference)
                     .await;
             }
         }
@@ -883,14 +895,16 @@ async fn write_inference(
 
 /// InferenceResponse and InferenceResultChunk determine what gets serialized and sent to the client
 
-#[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
+#[derive(Clone, Debug, PartialEq, Serialize, Deserialize, ts_rs::TS)]
+#[ts(export)]
 #[serde(untagged, rename_all = "snake_case")]
 pub enum InferenceResponse {
     Chat(ChatInferenceResponse),
     Json(JsonInferenceResponse),
 }
 
-#[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
+#[derive(Clone, Debug, PartialEq, Serialize, Deserialize, ts_rs::TS)]
+#[ts(export)]
 pub struct ChatInferenceResponse {
     pub inference_id: Uuid,
     pub episode_id: Uuid,
@@ -903,7 +917,8 @@ pub struct ChatInferenceResponse {
     pub finish_reason: Option<FinishReason>,
 }
 
-#[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
+#[derive(Clone, Debug, PartialEq, Serialize, Deserialize, ts_rs::TS)]
+#[ts(export)]
 pub struct JsonInferenceResponse {
     pub inference_id: Uuid,
     pub episode_id: Uuid,
@@ -1142,13 +1157,15 @@ pub struct InferenceModels<'a> {
 
 /// InferenceParams is the top-level struct for inference parameters.
 /// We backfill these from the configs given in the variants used and ultimately write them to the database.
-#[derive(Clone, Debug, Default, Deserialize, PartialEq, Serialize)]
+#[derive(Clone, Debug, Default, Deserialize, PartialEq, Serialize, ts_rs::TS)]
+#[ts(export)]
 #[serde(deny_unknown_fields)]
 pub struct InferenceParams {
     pub chat_completion: ChatCompletionInferenceParams,
 }
 
-#[derive(Clone, Debug, Default, Deserialize, PartialEq, Serialize)]
+#[derive(Clone, Debug, Default, Deserialize, PartialEq, Serialize, ts_rs::TS)]
+#[ts(export)]
 #[serde(deny_unknown_fields)]
 pub struct ChatCompletionInferenceParams {
     #[serde(skip_serializing_if = "Option::is_none")]
@@ -1205,6 +1222,79 @@ impl ChatCompletionInferenceParams {
     }
 }
 
+fn prepare_candidate_variants(
+    candidate_variants: &mut BTreeMap<String, Arc<VariantInfo>>,
+    tags: &mut HashMap<String, String>,
+    pinned_variant_name: Option<&str>,
+    dynamic_variant_config: Option<UninitializedVariantInfo>,
+    template_config: &mut Cow<'_, TemplateConfig>,
+    function: &FunctionConfig,
+    function_name: String,
+) -> Result<(), Error> {
+    match (pinned_variant_name, dynamic_variant_config) {
+        // If a variant is pinned, only that variant should be attempted
+        (Some(variant_name), None) => {
+            candidate_variants.retain(|k, _| k == variant_name);
+
+            // If the pinned variant doesn't exist, return an error
+            if candidate_variants.is_empty() {
+                return Err(ErrorDetails::UnknownVariant {
+                    name: variant_name.to_string(),
+                }
+                .into());
+            }
+            tags.insert(
+                "tensorzero::variant_pinned".to_string(),
+                variant_name.to_string(),
+            );
+        }
+        (None, Some(dynamic_variant_config)) => {
+            // Replace the variant config with just the dynamic variant
+            let candidate_variant_info = load_dynamic_variant_info(
+                dynamic_variant_config,
+                function.schemas(),
+                function_name,
+            )?;
+
+            // Replace templates in the template config with the ones passed in
+            // We Clone here so that we can still reference the old templates that don't conflict
+            let mut dynamic_template_config: TemplateConfig = template_config.clone().into_owned();
+            for path_with_contents in candidate_variant_info.get_all_template_paths() {
+                let template_name = path_with_contents.path.get_template_key();
+                if dynamic_template_config.contains_template(&template_name) {
+                    return Err(ErrorDetails::InvalidDynamicTemplatePath {
+                        name: template_name,
+                    }
+                    .into());
+                }
+                dynamic_template_config
+                    .add_template(template_name, path_with_contents.contents.clone())?;
+            }
+            *template_config = Cow::Owned(dynamic_template_config);
+            candidate_variants.clear();
+            candidate_variants.insert(
+                "tensorzero::dynamic_variant".to_string(),
+                Arc::new(candidate_variant_info),
+            );
+        }
+        (None, None) => {
+            // Remove all zero-weight variants - these can only be used if explicitly pinned above
+            candidate_variants.retain(|_, variant| {
+                // Retain 'None' and positive-weight variants, discarding zero-weight variants
+                variant.inner.weight().is_none_or(|w| w > 0.0)
+            });
+        }
+        _ => {
+            return Err(ErrorDetails::InvalidRequest {
+                message: "`variant_name` and `internal_dynamic_variant_config` cannot both be set."
+                    .to_string(),
+            }
+            .into())
+        }
+    }
+    Ok(())
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
