@@ -1,18 +1,12 @@
-use std::{
-    collections::HashMap,
-    fs::File,
-    io::{BufReader, Read},
-    path::Path,
-    sync::Arc,
-};
+use std::{collections::HashMap, sync::Arc};
 
 use serde::{Deserialize, Serialize};
 use tensorzero_derive::TensorZeroDeserialize;
 
 use crate::{
-    config_parser::{
+    config::{
         path::TomlRelativePath, MetricConfig, MetricConfigLevel, MetricConfigOptimize,
-        MetricConfigType, PathWithContents, TimeoutsConfig,
+        MetricConfigType, PathWithContents, SchemaData, TimeoutsConfig,
     },
     error::{Error, ErrorDetails},
     function::{FunctionConfig, FunctionConfigJson},
@@ -24,7 +18,7 @@ use crate::{
             BestOfNEvaluatorConfig as OnlineEvaluatorConfig, BestOfNSamplingConfig,
         },
         chain_of_thought::ChainOfThoughtConfig,
-        chat_completion::ChatCompletionConfig,
+        chat_completion::{ChatCompletionConfig, ChatTemplates, TemplateWithSchema},
         dicl::DiclConfig,
         mixture_of_n::{FuserConfig, MixtureOfNConfig},
         JsonMode, RetryConfig, VariantConfig, VariantInfo,
@@ -169,15 +163,15 @@ pub enum UninitializedEvaluationConfig {
 }
 
 impl UninitializedEvaluationConfig {
-    pub fn load<P: AsRef<Path>>(
+    pub fn load(
         self,
         functions: &HashMap<String, Arc<FunctionConfig>>,
-        base_path: P,
+
         evaluation_name: &str,
     ) -> EvaluationLoadResult {
         match self {
             UninitializedEvaluationConfig::Static(config) => {
-                config.load(functions, base_path, evaluation_name)
+                config.load(functions, evaluation_name)
             }
         }
     }
@@ -199,10 +193,9 @@ type EvaluationLoadResult = Result<
 >;
 
 impl UninitializedStaticEvaluationConfig {
-    pub fn load<P: AsRef<Path>>(
+    pub fn load(
         self,
         functions: &HashMap<String, Arc<FunctionConfig>>,
-        base_path: P,
         evaluation_name: &str,
     ) -> EvaluationLoadResult {
         if !functions.contains_key(&self.function_name) {
@@ -228,7 +221,7 @@ impl UninitializedStaticEvaluationConfig {
             .evaluators
             .into_iter()
             .map(|(name, config)| {
-                config.load(&base_path, evaluation_name, &name).map(
+                config.load(evaluation_name, &name).map(
                     |(evaluation_config, func_config, metric_config)| {
                         (name, evaluation_config, func_config, metric_config)
                     },
@@ -293,9 +286,8 @@ struct UninitializedLLMJudgeConfig {
 }
 
 impl UninitializedEvaluatorConfig {
-    pub fn load<P: AsRef<Path>>(
+    pub fn load(
         self,
-        base_path: &P,
         evaluation_name: &str,
         evaluator_name: &str,
     ) -> Result<(EvaluatorConfig, Option<FunctionConfig>, MetricConfig), Error> {
@@ -319,17 +311,43 @@ impl UninitializedEvaluatorConfig {
                 },
             )),
             UninitializedEvaluatorConfig::LLMJudge(params) => {
+                let user_schema_value: Option<serde_json::Value> = match params.input_format {
+                    LLMJudgeInputFormat::Serialized => Some(serde_json::from_str(LLM_JUDGE_USER_SCHEMA_TEXT)
+                        .map_err(|e| {
+                            Error::new(ErrorDetails::JsonSchema {
+                                message: format!("Failed to parse LLM judge user schema: {e}. This should never happen, please file a bug report at https://github.com/tensorzero/tensorzero/discussions/new?category=bug-reports."),
+                            })
+                        })?),
+                    LLMJudgeInputFormat::Messages => None,
+                };
+                let user_schema = user_schema_value
+                    .map(StaticJSONSchema::from_value)
+                    .transpose()?;
+                let output_schema_str = match params.output_type {
+                    LLMJudgeOutputType::Float => LLM_JUDGE_FLOAT_OUTPUT_SCHEMA_TEXT,
+                    LLMJudgeOutputType::Boolean => LLM_JUDGE_BOOLEAN_OUTPUT_SCHEMA_TEXT,
+                };
+                let output_schema_value = serde_json::from_str(output_schema_str)
+                    .map_err(|e| {
+                        Error::new(ErrorDetails::JsonSchema {
+                            message: format!("Failed to parse LLM judge output schema: {e}. This should never happen, please file a bug report at https://github.com/tensorzero/tensorzero/discussions/new?category=bug-reports."),
+                        })
+                    })?;
+                let output_schema = StaticJSONSchema::from_value(output_schema_value)?;
+                let implicit_tool_call_config =
+                    create_implicit_tool_call_config(output_schema.clone());
+
                 let mut variants = params
                     .variants
                     .into_iter()
                     .map(|(name, variant)| {
                         variant
                             .load(
-                                base_path,
                                 evaluation_name,
                                 evaluator_name,
                                 &params.input_format,
                                 &name,
+                                user_schema.clone(),
                             )
                             .map(|v| (name, v))
                     })
@@ -379,39 +397,17 @@ impl UninitializedEvaluatorConfig {
                         }
                     };
                 }
-                let user_schema_value: Option<serde_json::Value> = match params.input_format {
-                    LLMJudgeInputFormat::Serialized => Some(serde_json::from_str(LLM_JUDGE_USER_SCHEMA_TEXT)
-                        .map_err(|e| {
-                            Error::new(ErrorDetails::JsonSchema {
-                                message: format!("Failed to parse LLM judge user schema: {e}. This should never happen, please file a bug report at https://github.com/tensorzero/tensorzero/discussions/new?category=bug-reports."),
-                            })
-                        })?),
-                    LLMJudgeInputFormat::Messages => None,
-                };
-                let output_schema_str = match params.output_type {
-                    LLMJudgeOutputType::Float => LLM_JUDGE_FLOAT_OUTPUT_SCHEMA_TEXT,
-                    LLMJudgeOutputType::Boolean => LLM_JUDGE_BOOLEAN_OUTPUT_SCHEMA_TEXT,
-                };
-                let output_schema_value = serde_json::from_str(output_schema_str)
-                    .map_err(|e| {
-                        Error::new(ErrorDetails::JsonSchema {
-                            message: format!("Failed to parse LLM judge output schema: {e}. This should never happen, please file a bug report at https://github.com/tensorzero/tensorzero/discussions/new?category=bug-reports."),
-                        })
-                    })?;
-                let output_schema = StaticJSONSchema::from_value(&output_schema_value)?;
-                let implicit_tool_call_config =
-                    create_implicit_tool_call_config(output_schema.clone());
                 let variants = variants
                     .into_iter()
                     .map(|(name, variant)| (name, Arc::new(variant)))
                     .collect();
                 let function_config = FunctionConfig::Json(FunctionConfigJson {
                     variants,
-                    system_schema: None,
-                    user_schema: user_schema_value
-                        .map(|v| StaticJSONSchema::from_value(&v))
-                        .transpose()?,
-                    assistant_schema: None,
+                    schemas: SchemaData {
+                        system: None,
+                        user: user_schema,
+                        assistant: None,
+                    },
                     output_schema,
                     implicit_tool_call_config,
                     description: None,
@@ -484,41 +480,51 @@ struct UninitializedLLMJudgeChatCompletionVariantConfig {
 /// Converts a chat completion judge variant config to a chat completion config.
 /// This is factored out so that both the chain of thought and chat completion judges
 /// can use the same implementation.
-fn convert_chat_completion_judge_to_variant<P: AsRef<Path>>(
-    base_path: &P,
+fn convert_chat_completion_judge_to_variant(
     evaluation_name: &str,
     evaluator_name: &str,
     variant_name: &str,
     input_format: &LLMJudgeInputFormat,
     params: UninitializedLLMJudgeChatCompletionVariantConfig,
+    user_schema: Option<StaticJSONSchema>,
 ) -> Result<ChatCompletionConfig, Error> {
-    let system_instructions = read_system_instructions(&params.system_instructions, base_path)?;
+    let system_instructions = &params.system_instructions.read()?;
     let templated_system_instructions = format!(
         include_str!("llm_judge_system_instructions.txt"),
         system_instructions = system_instructions,
     );
-    let system_template_path =
-        get_template_path(evaluation_name, evaluator_name, variant_name, "system");
-    let system_template = PathWithContents {
-        // Not a real path but this is used as the handle everywhere as the content is already provided below
-        path: system_template_path.clone(),
-        contents: templated_system_instructions,
-    };
-    let user_template_path =
-        get_template_path(evaluation_name, evaluator_name, variant_name, "user");
+    let system_template_path = get_template_path(
+        evaluation_name,
+        evaluator_name,
+        variant_name,
+        "system",
+        templated_system_instructions,
+    );
+    let system_template = PathWithContents::from_path(system_template_path)?;
     let user_template = match input_format {
-        LLMJudgeInputFormat::Serialized => Some(PathWithContents {
-            path: user_template_path.clone(),
-            contents: include_str!("llm_judge_user_template.minijinja").to_string(),
-        }),
+        LLMJudgeInputFormat::Serialized => Some(PathWithContents::from_path(get_template_path(
+            evaluation_name,
+            evaluator_name,
+            variant_name,
+            "user",
+            include_str!("llm_judge_user_template.minijinja").to_string(),
+        ))?),
         LLMJudgeInputFormat::Messages => None,
     };
     Ok(ChatCompletionConfig {
         weight: get_weight(params.active),
         model: params.model,
-        system_template: Some(system_template),
-        user_template,
-        assistant_template: None,
+        templates: ChatTemplates {
+            system: Some(TemplateWithSchema {
+                template: system_template,
+                schema: None,
+            }),
+            user: user_template.map(|t| TemplateWithSchema {
+                template: t,
+                schema: user_schema,
+            }),
+            assistant: None,
+        },
         temperature: params.temperature,
         top_p: params.top_p,
         max_tokens: params.max_tokens,
@@ -594,10 +600,11 @@ fn get_template_path(
     evaluator_name: &str,
     variant_name: &str,
     template_name: &str,
+    data: String,
 ) -> TomlRelativePath {
     TomlRelativePath::new_fake_path(format!(
         "tensorzero::llm_judge::{evaluation_name}::{evaluator_name}::{variant_name}::{template_name}"
-    ))
+    ), data)
 }
 
 fn get_weight(active: Option<bool>) -> Option<f64> {
@@ -614,51 +621,48 @@ fn get_weight(active: Option<bool>) -> Option<f64> {
 }
 
 impl UninitializedLLMJudgeVariantInfo {
-    pub fn load<P: AsRef<Path>>(
+    pub fn load(
         self,
-        base_path: &P,
         evaluation_name: &str,
         evaluator_name: &str,
         input_format: &LLMJudgeInputFormat,
         variant_name: &str,
+        user_schema: Option<StaticJSONSchema>,
     ) -> Result<VariantInfo, Error> {
         let inner = match self.inner {
             UninitializedLLMJudgeVariantConfig::ChatCompletion(params) => {
                 VariantConfig::ChatCompletion(convert_chat_completion_judge_to_variant(
-                    base_path,
                     evaluation_name,
                     evaluator_name,
                     variant_name,
                     input_format,
                     params,
+                    user_schema,
                 )?)
             }
             UninitializedLLMJudgeVariantConfig::BestOfNSampling(params) => {
-                let evaluator_system_instructions =
-                    read_system_instructions(&params.evaluator.system_instructions, base_path)?;
+                let evaluator_system_instructions = &params.evaluator.system_instructions.read()?;
                 let templated_evaluator_system_instructions = format!(
                     include_str!("llm_judge_system_instructions.txt"),
                     system_instructions = evaluator_system_instructions,
                 );
-                let evaluator_system_template = PathWithContents {
-                    path: get_template_path(
-                        evaluation_name,
-                        evaluator_name,
-                        variant_name,
-                        "system",
-                    ),
-                    contents: templated_evaluator_system_instructions,
-                };
+                let evaluator_system_template = PathWithContents::from_path(get_template_path(
+                    evaluation_name,
+                    evaluator_name,
+                    variant_name,
+                    "system",
+                    templated_evaluator_system_instructions,
+                ))?;
                 let evaluator_user_template = match input_format {
-                    LLMJudgeInputFormat::Serialized => Some(PathWithContents {
-                        path: get_template_path(
+                    LLMJudgeInputFormat::Serialized => {
+                        Some(PathWithContents::from_path(get_template_path(
                             evaluation_name,
                             evaluator_name,
                             variant_name,
                             "user",
-                        ),
-                        contents: include_str!("llm_judge_user_template.minijinja").to_string(),
-                    }),
+                            include_str!("llm_judge_user_template.minijinja").to_string(),
+                        ))?)
+                    }
                     LLMJudgeInputFormat::Messages => None,
                 };
                 VariantConfig::BestOfNSampling(BestOfNSamplingConfig {
@@ -669,9 +673,17 @@ impl UninitializedLLMJudgeVariantInfo {
                         inner: ChatCompletionConfig {
                             weight: None,
                             model: params.evaluator.model,
-                            system_template: Some(evaluator_system_template),
-                            user_template: evaluator_user_template,
-                            assistant_template: None,
+                            templates: ChatTemplates {
+                                system: Some(TemplateWithSchema {
+                                    template: evaluator_system_template,
+                                    schema: None,
+                                }),
+                                user: evaluator_user_template.map(|t| TemplateWithSchema {
+                                    template: t,
+                                    schema: user_schema,
+                                }),
+                                assistant: None,
+                            },
                             temperature: params.evaluator.temperature,
                             top_p: params.evaluator.top_p,
                             max_tokens: params.evaluator.max_tokens,
@@ -688,31 +700,28 @@ impl UninitializedLLMJudgeVariantInfo {
                 })
             }
             UninitializedLLMJudgeVariantConfig::MixtureOfNSampling(params) => {
-                let fuser_system_instructions =
-                    read_system_instructions(&params.fuser.system_instructions, base_path)?;
+                let fuser_system_instructions = &params.fuser.system_instructions.read()?;
                 let templated_fuser_system_instructions = format!(
                     include_str!("llm_judge_system_instructions.txt"),
                     system_instructions = fuser_system_instructions,
                 );
-                let fuser_system_template = PathWithContents {
-                    path: get_template_path(
-                        evaluation_name,
-                        evaluator_name,
-                        variant_name,
-                        "system",
-                    ),
-                    contents: templated_fuser_system_instructions,
-                };
+                let fuser_system_template = PathWithContents::from_path(get_template_path(
+                    evaluation_name,
+                    evaluator_name,
+                    variant_name,
+                    "system",
+                    templated_fuser_system_instructions,
+                ))?;
                 let fuser_user_template = match input_format {
-                    LLMJudgeInputFormat::Serialized => Some(PathWithContents {
-                        path: get_template_path(
+                    LLMJudgeInputFormat::Serialized => {
+                        Some(PathWithContents::from_path(get_template_path(
                             evaluation_name,
                             evaluator_name,
                             variant_name,
                             "user",
-                        ),
-                        contents: include_str!("llm_judge_user_template.minijinja").to_string(),
-                    }),
+                            include_str!("llm_judge_user_template.minijinja").to_string(),
+                        ))?)
+                    }
                     LLMJudgeInputFormat::Messages => None,
                 };
                 VariantConfig::MixtureOfN(MixtureOfNConfig {
@@ -723,9 +732,17 @@ impl UninitializedLLMJudgeVariantInfo {
                         inner: ChatCompletionConfig {
                             weight: None,
                             model: params.fuser.model,
-                            system_template: Some(fuser_system_template),
-                            user_template: fuser_user_template,
-                            assistant_template: None,
+                            templates: ChatTemplates {
+                                system: Some(TemplateWithSchema {
+                                    template: fuser_system_template,
+                                    schema: None,
+                                }),
+                                user: fuser_user_template.map(|t| TemplateWithSchema {
+                                    template: t,
+                                    schema: user_schema,
+                                }),
+                                assistant: None,
+                            },
                             temperature: params.fuser.temperature,
                             top_p: params.fuser.top_p,
                             max_tokens: params.fuser.max_tokens,
@@ -744,7 +761,7 @@ impl UninitializedLLMJudgeVariantInfo {
             UninitializedLLMJudgeVariantConfig::Dicl(params) => {
                 let dicl_system_instructions = params
                     .system_instructions
-                    .map(|si| read_system_instructions(&si, base_path))
+                    .map(|si| si.read())
                     .transpose()?
                     .map(|si| {
                         format!(
@@ -775,12 +792,12 @@ impl UninitializedLLMJudgeVariantInfo {
             UninitializedLLMJudgeVariantConfig::ChainOfThought(params) => {
                 VariantConfig::ChainOfThought(ChainOfThoughtConfig {
                     inner: convert_chat_completion_judge_to_variant(
-                        base_path,
                         evaluation_name,
                         evaluator_name,
                         variant_name,
                         input_format,
                         params.inner,
+                        user_schema,
                     )?,
                 })
             }
@@ -792,33 +809,12 @@ impl UninitializedLLMJudgeVariantInfo {
     }
 }
 
-fn read_system_instructions<P: AsRef<Path>>(
-    path: &TomlRelativePath,
-    base_path: &P,
-) -> Result<String, Error> {
-    let path = base_path.as_ref().join(path.path());
-    let file = File::open(&path).map_err(|e| {
-        Error::new(ErrorDetails::FileRead {
-            message: format!("Failed to open system instructions file: {e}"),
-            file_path: path.to_string_lossy().to_string(),
-        })
-    })?;
-    let mut reader = BufReader::new(file);
-    let mut contents = String::new();
-    reader.read_to_string(&mut contents).map_err(|e| {
-        Error::new(ErrorDetails::FileRead {
-            message: format!("Failed to read system instructions file: {e}"),
-            file_path: path.to_string_lossy().to_string(),
-        })
-    })?;
-    Ok(contents)
-}
-
 /// NOTE: this function should not be called.
 /// In the code we already have a conversion from UninitializedLLMJudgeVariantConfig to VariantConfig.
 /// We want to make sure that there is an UninitializedLLMJudgeVariantConfig for each VariantConfig.
 /// This function should complain at compile time if we forget to update it when adding a new variant type.
 #[expect(dead_code)]
+#[expect(clippy::unnecessary_wraps)]
 fn check_convert_variant_to_llm_judge_variant(
     variant: VariantConfig,
 ) -> Result<UninitializedLLMJudgeVariantConfig, Error> {
@@ -828,7 +824,10 @@ fn check_convert_variant_to_llm_judge_variant(
                 UninitializedLLMJudgeChatCompletionVariantConfig {
                     active: Some(false),
                     model: variant.model,
-                    system_instructions: TomlRelativePath::new_fake_path(String::new()),
+                    system_instructions: TomlRelativePath::new_fake_path(
+                        String::new(),
+                        String::new(),
+                    ),
                     temperature: variant.temperature,
                     top_p: variant.top_p,
                     max_tokens: variant.max_tokens,
@@ -852,7 +851,10 @@ fn check_convert_variant_to_llm_judge_variant(
                     evaluator: UninitializedLLMJudgeChatCompletionVariantConfig {
                         active: Some(false),
                         model: variant.evaluator.inner.model,
-                        system_instructions: TomlRelativePath::new_fake_path(String::new()),
+                        system_instructions: TomlRelativePath::new_fake_path(
+                            String::new(),
+                            String::new(),
+                        ),
                         temperature: variant.evaluator.inner.temperature,
                         top_p: variant.evaluator.inner.top_p,
                         max_tokens: variant.evaluator.inner.max_tokens,
@@ -877,7 +879,10 @@ fn check_convert_variant_to_llm_judge_variant(
                     fuser: UninitializedLLMJudgeChatCompletionVariantConfig {
                         active: Some(false),
                         model: variant.fuser.inner.model,
-                        system_instructions: TomlRelativePath::new_fake_path(String::new()),
+                        system_instructions: TomlRelativePath::new_fake_path(
+                            String::new(),
+                            String::new(),
+                        ),
                         temperature: variant.fuser.inner.temperature,
                         top_p: variant.fuser.inner.top_p,
                         max_tokens: variant.fuser.inner.max_tokens,
@@ -919,7 +924,10 @@ fn check_convert_variant_to_llm_judge_variant(
                     inner: UninitializedLLMJudgeChatCompletionVariantConfig {
                         active: Some(false),
                         model: variant.inner.model,
-                        system_instructions: TomlRelativePath::new_fake_path(String::new()),
+                        system_instructions: TomlRelativePath::new_fake_path(
+                            String::new(),
+                            String::new(),
+                        ),
                         temperature: variant.inner.temperature,
                         top_p: variant.inner.top_p,
                         max_tokens: variant.inner.max_tokens,
@@ -945,37 +953,9 @@ mod tests {
     use std::path::PathBuf;
     use std::sync::Arc;
 
-    #[test]
-    fn test_read_system_instructions() {
-        let system_instructions = read_system_instructions(
-            &TomlRelativePath::new_for_tests(PathBuf::from(
-                "evaluations/evaluation1/llm_judge_bool/system_instructions.txt",
-            )),
-            &PathBuf::from("fixtures/config"),
-        )
-        .unwrap();
-        assert_eq!(
-            system_instructions,
-            "Return True if there is NSFW content in this generation.\n"
-        );
-
-        // Nonexistent file
-        let result = read_system_instructions(
-            &TomlRelativePath::new_for_tests(PathBuf::from(
-                "evaluations/evaluation1/llm_judge_bool/nonexistent.txt",
-            )),
-            &PathBuf::from("fixtures/config"),
-        );
-        assert_eq!(*result.unwrap_err().get_details(), ErrorDetails::FileRead {
-            message: "Failed to open system instructions file: No such file or directory (os error 2)".to_string(),
-            file_path: "fixtures/config/evaluations/evaluation1/llm_judge_bool/nonexistent.txt".to_string(),
-        });
-    }
-
     #[test]
     fn test_uninitialized_evaluation_config_load() {
         // Setup test fixtures
-        let base_path = PathBuf::from("fixtures/config");
         let evaluation_name = "test_evaluation";
 
         // Prepare function configs map with a function referenced in the evaluation
@@ -983,9 +963,7 @@ mod tests {
         let function_name = "generate_draft";
         let function_config = FunctionConfig::Json(FunctionConfigJson {
             variants: HashMap::new(),
-            system_schema: None,
-            user_schema: None,
-            assistant_schema: None,
+            schemas: SchemaData::default(),
             output_schema: create_test_schema(),
             implicit_tool_call_config: create_implicit_tool_call_config(create_test_schema()),
             description: None,
@@ -1005,7 +983,7 @@ mod tests {
                 function_name: function_name.to_string(),
             };
 
-            let result = uninitialized_config.load(&functions, &base_path, evaluation_name);
+            let result = uninitialized_config.load(&functions, evaluation_name);
             assert!(result.is_ok());
 
             let (config, additional_functions, metric_configs) = result.unwrap();
@@ -1047,7 +1025,7 @@ mod tests {
                             active: Some(true),
                             model: Arc::from("gpt-3.5-turbo"),
                             system_instructions:
-                                "evaluations/evaluation1/llm_judge_bool/system_instructions.txt"
+                                "fixtures/config/evaluations/evaluation1/llm_judge_bool/system_instructions.txt"
                                     .into(),
                             temperature: Some(0.7),
                             top_p: None,
@@ -1088,10 +1066,9 @@ mod tests {
                 function_name: function_name.to_string(),
             };
 
-            let result = uninitialized_config.load(&functions, &base_path, evaluation_name);
-            assert!(result.is_ok());
-
-            let (config, additional_functions, metric_configs) = result.unwrap();
+            let (config, additional_functions, metric_configs) = uninitialized_config
+                .load(&functions, evaluation_name)
+                .unwrap();
             assert_eq!(config.evaluators.len(), 1);
 
             // Verify LLM judge evaluator config
@@ -1118,8 +1095,8 @@ mod tests {
                 FunctionConfig::Json(json_config) => {
                     assert_eq!(json_config.variants.len(), 1);
                     assert!(json_config.variants.contains_key("test_variant"));
-                    assert!(json_config.system_schema.is_none());
-                    assert!(json_config.user_schema.is_some());
+                    assert!(json_config.schemas.system.is_none());
+                    assert!(json_config.schemas.user.is_some());
                     assert!(json_config.output_schema.value.is_object());
                 }
                 _ => panic!("Expected Json function config"),
@@ -1172,7 +1149,7 @@ mod tests {
                             active: Some(true),
                             model: Arc::from("gpt-3.5-turbo"),
                             system_instructions:
-                                "evaluations/evaluation1/llm_judge_bool/system_instructions.txt"
+                                "fixtures/config/evaluations/evaluation1/llm_judge_bool/system_instructions.txt"
                                     .into(),
                             temperature: Some(0.7),
                             top_p: None,
@@ -1213,10 +1190,9 @@ mod tests {
                 function_name: function_name.to_string(),
             };
 
-            let result = uninitialized_config.load(&functions, &base_path, evaluation_name);
-            assert!(result.is_ok());
-
-            let (config, additional_functions, metric_configs) = result.unwrap();
+            let (config, additional_functions, metric_configs) = uninitialized_config
+                .load(&functions, evaluation_name)
+                .unwrap();
             assert_eq!(config.evaluators.len(), 1);
 
             // Verify LLM judge evaluator config
@@ -1284,7 +1260,7 @@ mod tests {
                 function_name: "nonexistent_function".to_string(),
             };
 
-            let result = uninitialized_config.load(&functions, &base_path, evaluation_name);
+            let result = uninitialized_config.load(&functions, evaluation_name);
             assert!(result.is_err());
             assert!(matches!(
                 *result.unwrap_err().get_details(),
@@ -1305,8 +1281,7 @@ mod tests {
                 function_name: function_name.to_string(),
             };
 
-            let result =
-                uninitialized_config.load(&functions, &base_path, "invalid::evaluation::name");
+            let result = uninitialized_config.load(&functions, "invalid::evaluation::name");
             assert!(result.is_err());
             assert!(matches!(
                 *result.unwrap_err().get_details(),
@@ -1325,7 +1300,7 @@ mod tests {
                             active: Some(true),
                             model: Arc::from("gpt-3.5-turbo"),
                             system_instructions:
-                                "evaluations/evaluation1/llm_judge_bool/system_instructions.txt"
+                                "fixtures/config/evaluations/evaluation1/llm_judge_bool/system_instructions.txt"
                                     .into(),
                             temperature: Some(0.7),
                             top_p: None,
@@ -1353,8 +1328,8 @@ mod tests {
                             active: Some(true),
                             model: Arc::from("gpt-4"),
                             system_instructions: TomlRelativePath::new_for_tests(PathBuf::from(
-                                "evaluations/evaluation1/llm_judge_bool/system_instructions.txt",
-                            )),
+                                "fixtures/config/evaluations/evaluation1/llm_judge_bool/system_instructions.txt",
+                            ), None),
                             temperature: Some(0.5),
                             top_p: None,
                             max_tokens: Some(200),
@@ -1403,7 +1378,7 @@ mod tests {
                 function_name: function_name.to_string(),
             };
 
-            let result = uninitialized_config.load(&functions, &base_path, evaluation_name);
+            let result = uninitialized_config.load(&functions, evaluation_name);
             assert!(result.is_err());
             assert_eq!(
                 *result.unwrap_err().get_details(),
@@ -1415,7 +1390,6 @@ mod tests {
 
         // Test case 6: Error when evaluator name contains "::"
         {
-            let base_path = PathBuf::from(".");
             let evaluation_name = "test_evaluation";
             let function_name = "test_function";
 
@@ -1425,9 +1399,7 @@ mod tests {
                 Arc::new(FunctionConfig::Json(FunctionConfigJson {
                     variants: HashMap::new(),
                     output_schema: create_test_schema(),
-                    system_schema: None,
-                    user_schema: None,
-                    assistant_schema: None,
+                    schemas: SchemaData::default(),
                     implicit_tool_call_config: create_implicit_tool_call_config(
                         create_test_schema(),
                     ),
@@ -1446,7 +1418,7 @@ mod tests {
                 function_name: function_name.to_string(),
             };
 
-            let result = uninitialized_config.load(&functions, &base_path, evaluation_name);
+            let result = uninitialized_config.load(&functions, evaluation_name);
             assert!(result.is_err());
             assert_eq!(
                 *result.unwrap_err().get_details(),
@@ -1469,8 +1441,8 @@ mod tests {
                             active: Some(true),
                             model: Arc::from("gpt-3.5-turbo"),
                             system_instructions: TomlRelativePath::new_for_tests(PathBuf::from(
-                                "evaluations/evaluation1/llm_judge_bool/system_instructions.txt",
-                            )),
+                                "fixtures/config/evaluations/evaluation1/llm_judge_bool/system_instructions.txt",
+                            ), None),
                             temperature: Some(0.7),
                             top_p: None,
                             max_tokens: Some(100),
@@ -1510,7 +1482,7 @@ mod tests {
                 function_name: function_name.to_string(),
             };
 
-            let result = uninitialized_config.load(&functions, &base_path, evaluation_name);
+            let result = uninitialized_config.load(&functions, evaluation_name);
             assert!(result.is_ok());
 
             let (config, _additional_functions, _metric_configs) = result.unwrap();
@@ -1540,8 +1512,8 @@ mod tests {
                             active: None, // No 'active' field specified
                             model: Arc::from("gpt-3.5-turbo"),
                             system_instructions: TomlRelativePath::new_for_tests(PathBuf::from(
-                                "evaluations/evaluation1/llm_judge_bool/system_instructions.txt",
-                            )),
+                                "fixtures/config/evaluations/evaluation1/llm_judge_bool/system_instructions.txt",
+                            ), None),
                             temperature: Some(0.7),
                             top_p: None,
                             max_tokens: Some(100),
@@ -1579,7 +1551,7 @@ mod tests {
                 function_name: function_name.to_string(),
             };
 
-            let result = uninitialized_config.load(&functions, &base_path, evaluation_name);
+            let result = uninitialized_config.load(&functions, evaluation_name);
             assert!(result.is_ok());
 
             let (_config, additional_functions, _metric_configs) = result.unwrap();
@@ -1613,8 +1585,8 @@ mod tests {
                             active: Some(false), // Explicitly inactive
                             model: Arc::from("gpt-3.5-turbo"),
                             system_instructions: TomlRelativePath::new_for_tests(PathBuf::from(
-                                "evaluations/evaluation1/llm_judge_bool/system_instructions.txt",
-                            )),
+                                "fixtures/config/evaluations/evaluation1/llm_judge_bool/system_instructions.txt",
+                            ), None),
                             temperature: Some(0.7),
                             top_p: None,
                             max_tokens: Some(100),
@@ -1652,7 +1624,7 @@ mod tests {
                 function_name: function_name.to_string(),
             };
 
-            let result = uninitialized_config.load(&functions, &base_path, evaluation_name);
+            let result = uninitialized_config.load(&functions, evaluation_name);
             assert!(result.is_err());
             assert_eq!(
                 *result.unwrap_err().get_details(),
@@ -1674,6 +1646,6 @@ mod tests {
             },
             "required": ["result"]
         });
-        StaticJSONSchema::from_value(&schema_value).unwrap()
+        StaticJSONSchema::from_value(schema_value).unwrap()
     }
 }
