@@ -1,6 +1,5 @@
 use std::future::IntoFuture;
 use std::net::SocketAddr;
-use std::path::Path;
 use std::sync::Arc;
 
 use axum::extract::{rejection::JsonRejection, FromRequest, Json, Request};
@@ -8,14 +7,75 @@ use axum::routing::post;
 use axum::Router;
 use reqwest::{Client, Proxy};
 use serde::de::DeserializeOwned;
+use tokio::runtime::Handle;
 use tokio::sync::oneshot::Sender;
+use tokio_util::sync::CancellationToken;
 use tracing::instrument;
 
-use crate::clickhouse::migration_manager;
-use crate::clickhouse::ClickHouseConnectionInfo;
-use crate::config_parser::Config;
+use crate::config::{Config, ConfigFileGlob};
+use crate::db::clickhouse::migration_manager::{self, RunMigrationManagerArgs};
+use crate::db::clickhouse::ClickHouseConnectionInfo;
 use crate::endpoints;
 use crate::error::{Error, ErrorDetails};
+use crate::howdy::setup_howdy;
+
+/// Represents an active gateway (either standalone or embedded)
+/// The contained `app_state` can be freely cloned and dropped.
+/// However, dropping the `GatewayHandle` itself will wait for any
+/// needed background tasks to exit (in the future, this will
+/// include the ClickHouse batch insert task).
+///
+/// It's insufficient to put this kind of drop logic in `AppStateData` - since
+/// it can be freely cloned, any contained `Arc`s might only be dropped when
+/// the Tokio runtime is shutting down (e.g. if a background `tokio::spawn`
+/// task looped forever without using a `CancellationToken` to exit).
+/// During runtime shutdown, it's too late to call things like `tokio::spawn_blocking`,
+/// so we may be unable to safely wait for our batch insert task to finish writing.
+///
+/// `GatewayHandle` should *not* be wrapped in an `Arc` (or given a `Clone` impl),
+/// so that it's easy for us to tell where it gets dropped.
+///
+// Using `#[non_exhaustive]` has no effect within the crate
+#[expect(clippy::manual_non_exhaustive)]
+pub struct GatewayHandle {
+    pub app_state: AppStateData,
+    pub cancel_token: CancellationToken,
+    _private: (),
+}
+
+impl Drop for GatewayHandle {
+    fn drop(&mut self) {
+        self.cancel_token.cancel();
+        let handle = self
+            .app_state
+            .clickhouse_connection_info
+            .batcher_join_handle();
+        // Drop our `ClickHouseConnectionInfo`, so that we stop holding on to the `Arc<BatchSender>`
+        // This allows the batch writer task to exit (once all of the remaining `ClickhouseConnectionInfo`s are dropped)
+        self.app_state.clickhouse_connection_info = ClickHouseConnectionInfo::Disabled;
+        if let Some(handle) = handle {
+            tracing::info!("Waiting for ClickHouse batch writer to finish");
+            // This could block forever if:
+            // * We spawn a long-lived `tokio::task` that holds on to a `ClickhouseConnectionInfo`,
+            //   and isn't using our `CancellationToken` to exit.
+            // * The `GatewayHandle` is dropped from a task that's running other futures
+            //   concurrently (e.g. a `try_join_all` where one of the futures somehow drops a `GatewayHandle`).
+            //   In this case, the `block_in_place` call would prevent those futures from ever making progress,
+            //   causing a `ClickhouseConnectionInfo` (and therefore the `Arc<BatchSender>`) to never be dropped.
+            //   This is very unlikely, as we only create a `GatewayHandle` in a few places (the main gateway
+            //   and embedded client), and drop it when we're exiting.
+            //
+            // We err on the side of hanging the server on shutdown, rather than potentially exiting while
+            // we still have batched writes in-flight (or about to be written via an active `ClickhouseConnectionInfo`).
+            tokio::task::block_in_place(|| {
+                if let Err(e) = Handle::current().block_on(handle) {
+                    tracing::error!("Error in batch writer: {e}");
+                }
+            });
+            tracing::info!("ClickHouse batch writer finished");
+        }
+    }
+}
 
 /// State for the API
 #[derive(Clone)]
@@ -32,15 +92,15 @@ pub struct AppStateData {
 }
 pub type AppState = axum::extract::State<AppStateData>;
 
-impl AppStateData {
+impl GatewayHandle {
     pub async fn new(config: Arc<Config>) -> Result<Self, Error> {
-        let clickhouse_url = std::env::var("TENSORZERO_CLICKHOUSE_URL")
-            .ok()
-            .or_else(|| {
-                std::env::var("CLICKHOUSE_URL").ok().inspect(|_| {
-                    tracing::warn!("Deprecation Warning: The environment variable \"CLICKHOUSE_URL\" has been renamed to \"TENSORZERO_CLICKHOUSE_URL\" and will be removed in a future version. Please update your environment to use \"TENSORZERO_CLICKHOUSE_URL\" instead.");
-                })
-            });
+        let clickhouse_url = std::env::var("TENSORZERO_CLICKHOUSE_URL").ok();
+        if clickhouse_url.is_none()
+            && std::env::var("CLICKHOUSE_URL").is_ok()
+            && config.gateway.observability.enabled.is_none()
+        {
+            return Err(ErrorDetails::ClickHouseConfiguration { message: "`CLICKHOUSE_URL` is deprecated and no longer accepted. Please set `TENSORZERO_CLICKHOUSE_URL`".to_string() }.into());
+        }
         Self::new_with_clickhouse(config, clickhouse_url).await
     }
 
@@ -50,13 +110,18 @@ impl AppStateData {
     ) -> Result<Self, Error> {
         let clickhouse_connection_info = setup_clickhouse(&config, clickhouse_url, false).await?;
         let http_client = setup_http_client()?;
-
-        Ok(Self {
+        Ok(Self::new_with_clickhouse_and_http_client(
             config,
-            http_client,
             clickhouse_connection_info,
-            _private: (),
-        })
+            http_client,
+        ))
+    }
+
+    #[cfg(test)]
+    pub fn new_unit_test_data(config: Arc<Config>, clickhouse_healthy: bool) -> Self {
+        let http_client = reqwest::Client::new();
+        let clickhouse_connection_info = ClickHouseConnectionInfo::new_mock(clickhouse_healthy);
+        Self::new_with_clickhouse_and_http_client(config, clickhouse_connection_info, http_client)
     }
 
     pub fn new_with_clickhouse_and_http_client(
@@ -64,22 +129,20 @@ impl AppStateData {
         clickhouse_connection_info: ClickHouseConnectionInfo,
         http_client: Client,
     ) -> Self {
+        let cancel_token = CancellationToken::new();
+        setup_howdy(
+            &config,
+            clickhouse_connection_info.clone(),
+            cancel_token.clone(),
+        );
         Self {
-            config,
-            http_client,
-            clickhouse_connection_info,
-            _private: (),
-        }
-    }
-
-    #[cfg(test)]
-    pub fn new_unit_test_data(config: Arc<Config>, clickhouse_healthy: bool) -> Self {
-        let http_client = reqwest::Client::new();
-        let clickhouse_connection_info = ClickHouseConnectionInfo::new_mock(clickhouse_healthy);
-        AppStateData {
-            config,
-            http_client,
-            clickhouse_connection_info,
+            app_state: AppStateData {
+                config,
+                http_client,
+                clickhouse_connection_info,
+                _private: (),
+            },
+            cancel_token,
             _private: (),
         }
     }
@@ -105,7 +168,11 @@ pub async fn setup_clickhouse(
         }
         // Observability enabled and ClickHouse URL provided
         (Some(true), Some(clickhouse_url)) => {
-            ClickHouseConnectionInfo::new(&clickhouse_url).await?
+            ClickHouseConnectionInfo::new(
+                &clickhouse_url,
+                config.gateway.observability.batch_writes.clone(),
+            )
+            .await?
         }
         // Observability default and no ClickHouse URL
         (None, None) => {
@@ -118,12 +185,23 @@ pub async fn setup_clickhouse(
             ClickHouseConnectionInfo::new_disabled()
         }
         // Observability default and ClickHouse URL provided
-        (None, Some(clickhouse_url)) => ClickHouseConnectionInfo::new(&clickhouse_url).await?,
+        (None, Some(clickhouse_url)) => {
+            ClickHouseConnectionInfo::new(
+                &clickhouse_url,
+                config.gateway.observability.batch_writes.clone(),
+            )
+            .await?
+        }
     };
 
     // Run ClickHouse migrations (if any) if we have a production ClickHouse connection
     if let ClickHouseConnectionInfo::Production { .. } = &clickhouse_connection_info {
-        migration_manager::run(&clickhouse_connection_info).await?;
+        migration_manager::run(RunMigrationManagerArgs {
+            clickhouse: &clickhouse_connection_info,
+            skip_completed_migrations: true,
+            manual_run: false,
+        })
+        .await?;
     }
     Ok(clickhouse_connection_info)
 }
@@ -200,9 +278,12 @@ pub fn setup_http_client() -> Result<Client, Error> {
     })
 }
 
+// We hold on to these fields so that their Drop impls run when `ShutdownHandle` is dropped
 pub struct ShutdownHandle {
     #[expect(dead_code)]
     sender: Sender<()>,
+    #[expect(dead_code)]
+    gateway_handle: GatewayHandle,
 }
 
 /// Starts a new HTTP TensorZero gateway on an unused port, with only the openai-compatible endpoint enabled.
@@ -229,19 +310,26 @@ pub async fn start_openai_compatible_gateway(
     })?;
 
     let config = if let Some(config_file) = config_file {
-        Arc::new(Config::load_and_verify_from_path(Path::new(&config_file)).await?)
+        Arc::new(Config::load_and_verify_from_path(&ConfigFileGlob::new(config_file)?).await?)
     } else {
         Arc::new(Config::default())
     };
-    let app_state = AppStateData::new_with_clickhouse(config, clickhouse_url).await?;
+    let gateway_handle = GatewayHandle::new_with_clickhouse(config, clickhouse_url).await?;
 
+    // TODO(# 3191): Implement a trait for openai compatible endpoints
+    // so this logic can be centralized to one place and not reimplemented in
+    // our gateway main.
     let router = Router::new()
         .route(
             "/openai/v1/chat/completions",
             post(endpoints::openai_compatible::inference_handler),
         )
+        .route(
+            "/openai/v1/embeddings",
+            post(endpoints::openai_compatible::embeddings_handler),
+        )
         .fallback(endpoints::fallback::handle_404)
-        .with_state(app_state);
+        .with_state(gateway_handle.app_state.clone());
 
     let (sender, recv) = tokio::sync::oneshot::channel::<()>();
     let shutdown_fut = async move {
@@ -253,7 +341,13 @@ pub async fn start_openai_compatible_gateway(
             .with_graceful_shutdown(shutdown_fut)
             .into_future(),
     );
-    Ok((bind_addr, ShutdownHandle { sender }))
+    Ok((
+        bind_addr,
+        ShutdownHandle {
+            sender,
+            gateway_handle,
+        },
+    ))
 }
 
 #[cfg(test)]
@@ -261,7 +355,7 @@ mod tests {
     use tracing_test::traced_test;
 
     use super::*;
-    use crate::config_parser::{gateway::GatewayConfig, ObservabilityConfig};
+    use crate::config::{gateway::GatewayConfig, ObservabilityConfig};
 
     #[tokio::test]
     #[traced_test]
@@ -271,6 +365,7 @@ mod tests {
             observability: ObservabilityConfig {
                 enabled: Some(false),
                 async_writes: false,
+                batch_writes: Default::default(),
             },
             bind_address: None,
             debug: false,
@@ -278,6 +373,8 @@ mod tests {
             export: Default::default(),
             base_path: None,
             unstable_error_json: false,
+            unstable_disable_feedback_target_validation: false,
+            disable_pseudonymous_usage_analytics: false,
         };
 
         let config = Box::leak(Box::new(Config {
@@ -299,6 +396,7 @@ mod tests {
             observability: ObservabilityConfig {
                 enabled: None,
                 async_writes: false,
+                batch_writes: Default::default(),
             },
             unstable_error_json: false,
             ..Default::default()
@@ -325,6 +423,7 @@ mod tests {
             observability: ObservabilityConfig {
                 enabled: Some(true),
                 async_writes: false,
+                batch_writes: Default::default(),
             },
             bind_address: None,
             debug: false,
@@ -332,6 +431,8 @@ mod tests {
             export: Default::default(),
             base_path: None,
             unstable_error_json: false,
+            unstable_disable_feedback_target_validation: false,
+            disable_pseudonymous_usage_analytics: false,
         };
 
         let config = Box::leak(Box::new(Config {
@@ -349,6 +450,7 @@ mod tests {
             observability: ObservabilityConfig {
                 enabled: Some(true),
                 async_writes: false,
+                batch_writes: Default::default(),
             },
             bind_address: None,
             debug: false,
@@ -356,6 +458,8 @@ mod tests {
             export: Default::default(),
             base_path: None,
             unstable_error_json: false,
+            unstable_disable_feedback_target_validation: false,
+            disable_pseudonymous_usage_analytics: false,
         };
         let config = Box::leak(Box::new(Config {
             gateway: gateway_config,
@@ -375,6 +479,7 @@ mod tests {
             observability: ObservabilityConfig {
                 enabled: Some(true),
                 async_writes: false,
+                batch_writes: Default::default(),
             },
             bind_address: None,
             debug: false,
@@ -382,6 +487,8 @@ mod tests {
             export: Default::default(),
             base_path: None,
             unstable_error_json: false,
+            unstable_disable_feedback_target_validation: false,
+            disable_pseudonymous_usage_analytics: false,
         };
         let config = Config {
             gateway: gateway_config,
