@@ -3,8 +3,8 @@
 #![expect(async_fn_in_trait)]
 
 pub mod cache;
-pub mod clickhouse;
-pub mod config_parser; // TensorZero config file
+pub mod config; // TensorZero config file
+pub mod db;
 pub mod embeddings; // embedding inference
 pub mod endpoints; // API endpoints
 pub mod error; // error handling
