@@ -303,7 +303,7 @@ enum DeepSeekResponseFormat {
 }
 
 impl DeepSeekResponseFormat {
-    fn new(json_mode: &ModelInferenceRequestJsonMode) -> Option<Self> {
+    fn new(json_mode: ModelInferenceRequestJsonMode) -> Option<Self> {
         match json_mode {
             ModelInferenceRequestJsonMode::On => Some(DeepSeekResponseFormat::JsonObject),
             // For now, we never explicitly send `DeepSeekResponseFormat::Text`
@@ -371,7 +371,7 @@ impl<'a> DeepSeekRequest<'a> {
             tracing::warn!("DeepSeek provider does not support strict JSON mode. Downgrading to normal JSON mode.");
         }
 
-        let response_format = DeepSeekResponseFormat::new(&request.json_mode);
+        let response_format = DeepSeekResponseFormat::new(request.json_mode);
 
         // NOTE: as mentioned by the DeepSeek team here: https://github.com/deepseek-ai/DeepSeek-R1?tab=readme-ov-file#usage-recommendations
         // the R1 series of models does not perform well with the system prompt. As we move towards first-class support for reasoning models we should check
@@ -589,7 +589,7 @@ pub(super) fn prepare_deepseek_messages<'a>(
     model_name: &'a str,
 ) -> Result<Vec<OpenAIRequestMessage<'a>>, Error> {
     let mut messages = Vec::with_capacity(request.messages.len());
-    for message in request.messages.iter() {
+    for message in &request.messages {
         messages.extend(tensorzero_to_openai_messages(message, PROVIDER_TYPE)?);
     }
     // If this is an R1 model, prepend the system message as the first user message instead of using it as a system message
