@@ -531,11 +531,11 @@ pub(super) fn prepare_groq_messages<'a>(
     request: &'a ModelInferenceRequest<'_>,
 ) -> Result<Vec<GroqRequestMessage<'a>>, Error> {
     let mut messages = Vec::with_capacity(request.messages.len());
-    for message in request.messages.iter() {
+    for message in &request.messages {
         messages.extend(tensorzero_to_groq_messages(message)?);
     }
     if let Some(system_msg) =
-        tensorzero_to_groq_system_message(request.system.as_deref(), &request.json_mode, &messages)
+        tensorzero_to_groq_system_message(request.system.as_deref(), request.json_mode, &messages)
     {
         messages.insert(0, system_msg);
     }
@@ -571,7 +571,7 @@ pub(super) fn prepare_groq_tools<'a>(
 /// So, we need to format the instructions to include "Respond using JSON." if it doesn't already.
 pub(super) fn tensorzero_to_groq_system_message<'a>(
     system: Option<&'a str>,
-    json_mode: &ModelInferenceRequestJsonMode,
+    json_mode: ModelInferenceRequestJsonMode,
     messages: &[GroqRequestMessage<'a>],
 ) -> Option<GroqRequestMessage<'a>> {
     match system {
@@ -601,7 +601,7 @@ pub(super) fn tensorzero_to_groq_system_message<'a>(
             }
             .into()
         }
-        None => match *json_mode {
+        None => match json_mode {
             ModelInferenceRequestJsonMode::On => {
                 Some(GroqRequestMessage::System(GroqSystemRequestMessage {
                     content: Cow::Owned("Respond using JSON.".to_string()),
@@ -629,7 +629,7 @@ fn tensorzero_to_groq_user_messages(
     let mut messages = Vec::new();
     let mut user_content_blocks = Vec::new();
 
-    for block in content_blocks.iter() {
+    for block in content_blocks {
         match block {
             ContentBlock::Text(Text { text }) => {
                 user_content_blocks.push(GroqContentBlock::Text {
@@ -691,7 +691,7 @@ fn tensorzero_to_groq_assistant_messages(
     let mut assistant_content_blocks = Vec::new();
     let mut assistant_tool_calls = Vec::new();
 
-    for block in content_blocks.iter() {
+    for block in content_blocks {
         match block {
             ContentBlock::Text(Text { text }) => {
                 assistant_content_blocks.push(GroqContentBlock::Text {
@@ -773,7 +773,7 @@ enum GroqResponseFormat<'a> {
 }
 
 impl<'a> GroqResponseFormat<'a> {
-    fn new(json_mode: &ModelInferenceRequestJsonMode, output_schema: Option<&'a Value>) -> Self {
+    fn new(json_mode: ModelInferenceRequestJsonMode, output_schema: Option<&'a Value>) -> Self {
         match json_mode {
             ModelInferenceRequestJsonMode::On | ModelInferenceRequestJsonMode::Strict => {
                 GroqResponseFormat::JsonObject {
@@ -914,7 +914,7 @@ impl<'a> GroqRequest<'a> {
         request: &'a ModelInferenceRequest<'_>,
     ) -> Result<GroqRequest<'a>, Error> {
         let response_format = Some(GroqResponseFormat::new(
-            &request.json_mode,
+            request.json_mode,
             request.output_schema,
         ));
         let stream_options = if request.stream {
@@ -2162,17 +2162,17 @@ mod tests {
         // Test JSON mode On
         let json_mode = ModelInferenceRequestJsonMode::On;
         let output_schema = None;
-        let format = GroqResponseFormat::new(&json_mode, output_schema);
+        let format = GroqResponseFormat::new(json_mode, output_schema);
         assert_eq!(format, GroqResponseFormat::JsonObject { schema: None });
 
         // Test JSON mode Off
         let json_mode = ModelInferenceRequestJsonMode::Off;
-        let format = GroqResponseFormat::new(&json_mode, output_schema);
+        let format = GroqResponseFormat::new(json_mode, output_schema);
         assert_eq!(format, GroqResponseFormat::Text);
 
         // Test JSON mode Strict with no schema
         let json_mode = ModelInferenceRequestJsonMode::Strict;
-        let format = GroqResponseFormat::new(&json_mode, output_schema);
+        let format = GroqResponseFormat::new(json_mode, output_schema);
         assert_eq!(format, GroqResponseFormat::JsonObject { schema: None });
 
         // Test JSON mode Strict with schema
@@ -2184,7 +2184,7 @@ mod tests {
             }
         });
         let output_schema = Some(&json_schema);
-        let format = GroqResponseFormat::new(&json_mode, output_schema);
+        let format = GroqResponseFormat::new(json_mode, output_schema);
         assert_eq!(
             format,
             GroqResponseFormat::JsonObject {
@@ -2199,7 +2199,7 @@ mod tests {
         let system = None;
         let json_mode = ModelInferenceRequestJsonMode::Off;
         let messages: Vec<GroqRequestMessage> = vec![];
-        let result = tensorzero_to_groq_system_message(system, &json_mode, &messages);
+        let result = tensorzero_to_groq_system_message(system, json_mode, &messages);
         assert_eq!(result, None);
 
         // Test Case 2: system is Some, json_mode is On, messages contain "json"
@@ -2221,7 +2221,7 @@ mod tests {
         let expected = Some(GroqRequestMessage::System(GroqSystemRequestMessage {
             content: Cow::Borrowed("System instructions"),
         }));
-        let result = tensorzero_to_groq_system_message(system, &json_mode, &messages);
+        let result = tensorzero_to_groq_system_message(system, json_mode, &messages);
         assert_eq!(result, expected);
 
         // Test Case 3: system is Some, json_mode is On, messages do not contain "json"
@@ -2244,7 +2244,7 @@ mod tests {
         let expected = Some(GroqRequestMessage::System(GroqSystemRequestMessage {
             content: Cow::Owned(expected_content),
         }));
-        let result = tensorzero_to_groq_system_message(system, &json_mode, &messages);
+        let result = tensorzero_to_groq_system_message(system, json_mode, &messages);
         assert_eq!(result, expected);
 
         // Test Case 4: system is Some, json_mode is Off
@@ -2266,7 +2266,7 @@ mod tests {
         let expected = Some(GroqRequestMessage::System(GroqSystemRequestMessage {
             content: Cow::Borrowed("System instructions"),
         }));
-        let result = tensorzero_to_groq_system_message(system, &json_mode, &messages);
+        let result = tensorzero_to_groq_system_message(system, json_mode, &messages);
         assert_eq!(result, expected);
 
         // Test Case 5: system is Some, json_mode is Strict
@@ -2288,7 +2288,7 @@ mod tests {
         let expected = Some(GroqRequestMessage::System(GroqSystemRequestMessage {
             content: Cow::Borrowed("System instructions"),
         }));
-        let result = tensorzero_to_groq_system_message(system, &json_mode, &messages);
+        let result = tensorzero_to_groq_system_message(system, json_mode, &messages);
         assert_eq!(result, expected);
 
         // Test Case 6: system contains "json", json_mode is On
@@ -2302,7 +2302,7 @@ mod tests {
         let expected = Some(GroqRequestMessage::System(GroqSystemRequestMessage {
             content: Cow::Borrowed("Respond using JSON.\n\nSystem instructions"),
         }));
-        let result = tensorzero_to_groq_system_message(system, &json_mode, &messages);
+        let result = tensorzero_to_groq_system_message(system, json_mode, &messages);
         assert_eq!(result, expected);
 
         // Test Case 7: system is None, json_mode is On
@@ -2324,7 +2324,7 @@ mod tests {
         let expected = Some(GroqRequestMessage::System(GroqSystemRequestMessage {
             content: Cow::Owned("Respond using JSON.".to_string()),
         }));
-        let result = tensorzero_to_groq_system_message(system, &json_mode, &messages);
+        let result = tensorzero_to_groq_system_message(system, json_mode, &messages);
         assert_eq!(result, expected);
 
         // Test Case 8: system is None, json_mode is Strict
@@ -2344,7 +2344,7 @@ mod tests {
             }),
         ];
 
-        let result = tensorzero_to_groq_system_message(system, &json_mode, &messages);
+        let result = tensorzero_to_groq_system_message(system, json_mode, &messages);
         assert!(result.is_none());
 
         // Test Case 9: system is None, json_mode is On, with empty messages
@@ -2354,7 +2354,7 @@ mod tests {
         let expected = Some(GroqRequestMessage::System(GroqSystemRequestMessage {
             content: Cow::Owned("Respond using JSON.".to_string()),
         }));
-        let result = tensorzero_to_groq_system_message(system, &json_mode, &messages);
+        let result = tensorzero_to_groq_system_message(system, json_mode, &messages);
         assert_eq!(result, expected);
 
         // Test Case 10: system is None, json_mode is Off, with messages containing "json"
@@ -2366,7 +2366,7 @@ mod tests {
             }],
         })];
         let expected = None;
-        let result = tensorzero_to_groq_system_message(system, &json_mode, &messages);
+        let result = tensorzero_to_groq_system_message(system, json_mode, &messages);
         assert_eq!(result, expected);
     }
 
