@@ -1,13 +1,12 @@
-use std::fs;
-use std::path::Path;
 use std::sync::Arc;
 
 use serde::Deserialize;
 use serde::Serialize;
 
-use crate::config_parser::path::TomlRelativePath;
-use crate::config_parser::LoadableConfig;
-use crate::config_parser::PathWithContents;
+use crate::config::path::TomlRelativePath;
+use crate::config::LoadableConfig;
+use crate::config::PathWithContents;
+use crate::embeddings::EmbeddingEncodingFormat;
 use crate::embeddings::{EmbeddingModelTable, EmbeddingResponseWithMetadata};
 use crate::endpoints::inference::InferenceModels;
 use crate::inference::types::extra_body::{ExtraBodyConfig, FullExtraBodyConfig};
@@ -23,7 +22,7 @@ use crate::model_table::ShorthandModelConfig;
 use crate::{
     embeddings::EmbeddingRequest,
     endpoints::inference::{InferenceClients, InferenceParams},
-    error::{Error, ErrorDetails},
+    error::{Error, ErrorDetails, IMPOSSIBLE_ERROR_MESSAGE},
     function::FunctionConfig,
     inference::types::{
         ContentBlockChatOutput, InferenceResult, InferenceResultStream, JsonInferenceOutput,
@@ -57,12 +56,15 @@ pub struct DiclConfig {
     pub max_tokens: Option<u32>,
     pub seed: Option<u32>,
     pub json_mode: Option<JsonMode>,
+    #[cfg_attr(test, ts(skip))]
     pub extra_body: Option<ExtraBodyConfig>,
+    #[cfg_attr(test, ts(skip))]
     pub extra_headers: Option<ExtraHeadersConfig>,
     pub retries: RetryConfig,
 }
 
-#[derive(Debug, Deserialize)]
+#[derive(Clone, Debug, Deserialize, Serialize, ts_rs::TS)]
+#[ts(export)]
 #[serde(deny_unknown_fields)]
 pub struct UninitializedDiclConfig {
     #[serde(default)]
@@ -80,10 +82,12 @@ pub struct UninitializedDiclConfig {
     pub seed: Option<u32>,
     pub json_mode: Option<JsonMode>,
     #[serde(default)]
+    #[ts(skip)]
     pub extra_body: Option<ExtraBodyConfig>,
     #[serde(default)]
     pub retries: RetryConfig,
     #[serde(default)]
+    #[ts(skip)]
     pub extra_headers: Option<ExtraHeadersConfig>,
 }
 
@@ -93,7 +97,7 @@ impl Variant for DiclConfig {
         input: &ResolvedInput,
         models: &'request InferenceModels<'a>,
         function: &'a FunctionConfig,
-        inference_config: &'request InferenceConfig<'static, 'request>,
+        inference_config: &'request InferenceConfig<'request>,
         clients: &'request InferenceClients<'request>,
         inference_params: InferenceParams,
     ) -> Result<InferenceResult, Error> {
@@ -107,11 +111,7 @@ impl Variant for DiclConfig {
                 models.embedding_models,
                 clients,
                 inference_config.function_name,
-                inference_config.variant_name.ok_or_else(|| {
-                    Error::new(ErrorDetails::InvalidDiclConfig {
-                        message: "missing variant_name".to_string(),
-                    })
-                })?,
+                inference_config.variant_name,
                 function,
             )
             .await?;
@@ -150,7 +150,9 @@ impl Variant for DiclConfig {
         // Add the embedding to the model inference results
         inference_response
             .mut_model_inference_results()
-            .push(embedding_response.into());
+            // This can only fail if the embedding response has a request with > 1 input,
+            // which should never happen as we only send one input at a time
+            .push(embedding_response.try_into()?);
         Ok(inference_response)
     }
 
@@ -159,7 +161,7 @@ impl Variant for DiclConfig {
         input: &ResolvedInput,
         models: &'request InferenceModels<'_>,
         function: &FunctionConfig,
-        inference_config: &'request InferenceConfig<'static, 'request>,
+        inference_config: &'request InferenceConfig<'request>,
         clients: &'request InferenceClients<'request>,
         inference_params: InferenceParams,
     ) -> Result<(InferenceResultStream, ModelUsedInfo), Error> {
@@ -173,11 +175,7 @@ impl Variant for DiclConfig {
                 models.embedding_models,
                 clients,
                 inference_config.function_name,
-                inference_config.variant_name.ok_or_else(|| {
-                    Error::new(ErrorDetails::InvalidDiclConfig {
-                        message: "missing variant_name".to_string(),
-                    })
-                })?,
+                inference_config.variant_name,
                 function,
             )
             .await?;
@@ -212,7 +210,7 @@ impl Variant for DiclConfig {
         // Add the embedding to the model inference results
         model_used_info
             .previous_model_inference_results
-            .push(embedding_response.into());
+            .push(embedding_response.try_into()?);
         Ok((inference_result_stream, model_used_info))
     }
 
@@ -271,7 +269,7 @@ impl Variant for DiclConfig {
         _input: &[ResolvedInput],
         _models: &'a InferenceModels<'a>,
         _function: &'a FunctionConfig,
-        _inference_configs: &'a [InferenceConfig<'a, 'a>],
+        _inference_configs: &'a [InferenceConfig<'a>],
         _clients: &'a InferenceClients<'a>,
         _inference_params: Vec<InferenceParams>,
     ) -> Result<StartBatchModelInferenceWithMetadata<'a>, Error> {
@@ -334,7 +332,9 @@ impl DiclConfig {
             })?;
 
         let embedding_request = EmbeddingRequest {
-            input: serialized_input.to_string(),
+            input: serialized_input.into(),
+            dimensions: None,
+            encoding_format: EmbeddingEncodingFormat::Float,
         };
 
         // Embed the input via an API request
@@ -345,24 +345,35 @@ impl DiclConfig {
         // Wrap the embedding in a response with metadata
         let embedding_response_with_metadata =
             EmbeddingResponseWithMetadata::new(embedding_response, self.embedding_model.clone());
+        let [embedding_vector] = embedding_response_with_metadata.embeddings.as_slice() else {
+            return Err(ErrorDetails::InternalError {
+                message: format!(
+                    "Embedding model returned multiple vectors. {IMPOSSIBLE_ERROR_MESSAGE}"
+                ),
+            }
+            .into());
+        };
 
         // Format the embedding as a string for ClickHouse
         let formatted_embedding = format!(
             "[{}]",
-            embedding_response_with_metadata
-                .embedding
+            embedding_vector
+                .as_float()
+                .ok_or_else(|| Error::new(ErrorDetails::InternalError {
+                    message: format!("Failed to convert DICL embedding to float array. {IMPOSSIBLE_ERROR_MESSAGE}")
+                }))?
                 .iter()
                 .map(|&x| x.to_string())
                 .collect::<Vec<_>>()
                 .join(",")
         );
         let query = format!(
-            r#"SELECT input, output, cosineDistance(embedding, {}) as distance
+            r"SELECT input, output, cosineDistance(embedding, {}) as distance
                    FROM DynamicInContextLearningExample
                    WHERE function_name='{}' AND variant_name='{}'
                    ORDER BY distance ASC
                    LIMIT {}
-                   FORMAT JSONEachRow"#,
+                   FORMAT JSONEachRow",
             formatted_embedding, function_name, variant_name, self.k
         );
 
@@ -403,7 +414,7 @@ impl DiclConfig {
     /// The second message is an Assistant message with the output as native output blocks
     ///   - For chat messages, this is a simple vector of ContentBlocks
     ///   - For JSON messages, this is a single JSON output block (as Text)
-    fn prepare_message(&self, example: &Example) -> Result<Vec<RequestMessage>, Error> {
+    fn prepare_message(example: &Example) -> Result<Vec<RequestMessage>, Error> {
         let mut messages = Vec::new();
         let input = match example {
             Example::Chat(chat_example) => chat_example.input.clone(),
@@ -445,7 +456,7 @@ impl DiclConfig {
         Ok(messages)
     }
 
-    fn prepare_input_message(&self, input: &ResolvedInput) -> Result<RequestMessage, Error> {
+    fn prepare_input_message(input: &ResolvedInput) -> Result<RequestMessage, Error> {
         let content = vec![serde_json::to_string(&input)
             .map_err(|e| {
                 Error::new(ErrorDetails::Serialization {
@@ -466,7 +477,7 @@ impl DiclConfig {
         input: &ResolvedInput,
         examples: &[Example],
         function: &'a FunctionConfig,
-        inference_config: &'request InferenceConfig<'a, 'request>,
+        inference_config: &'request InferenceConfig<'request>,
         stream: bool,
         inference_params: &mut InferenceParams,
     ) -> Result<ModelInferenceRequest<'request>, Error>
@@ -497,11 +508,11 @@ impl DiclConfig {
         }
         let messages = examples
             .iter()
-            .map(|example| self.prepare_message(example))
+            .map(Self::prepare_message)
             .collect::<Result<Vec<Vec<RequestMessage>>, _>>()?
             .into_iter()
             .flatten()
-            .chain(std::iter::once(self.prepare_input_message(input)?))
+            .chain(std::iter::once(Self::prepare_input_message(input)?))
             .collect::<Vec<_>>();
 
         let system = Some(self.system_instructions.clone());
@@ -533,6 +544,7 @@ impl DiclConfig {
             inference_extra_headers: inference_config
                 .extra_headers
                 .clone()
+                .into_owned()
                 .filter(inference_config.variant_name),
         };
         prepare_model_inference_request(
@@ -612,19 +624,9 @@ pub fn default_system_instructions() -> String {
 }
 
 impl LoadableConfig<DiclConfig> for UninitializedDiclConfig {
-    /// Since the system instructions are optional and may be a path to a file,
-    /// we need to load them here so that we can use the base_path to resolve
-    /// any relative paths.
-    fn load<P: AsRef<Path>>(self, base_path: P) -> Result<DiclConfig, Error> {
+    fn load(self) -> Result<DiclConfig, Error> {
         let system_instructions = match self.system_instructions {
-            Some(path) => {
-                let path = base_path.as_ref().join(path.path());
-                fs::read_to_string(path).map_err(|e| {
-                    Error::new(ErrorDetails::Config {
-                        message: format!("Failed to read system instructions: {e}"),
-                    })
-                })?
-            }
+            Some(path) => path.read()?,
             None => default_system_instructions(),
         };
 
@@ -665,9 +667,6 @@ mod tests {
 
     #[test]
     fn test_prepare_message() {
-        // Create an instance of DiclConfig (assuming default implementation is available)
-        let dicl_config = DiclConfig::default();
-
         // ---------- Test with ChatExample ----------
 
         // Mock Input data
@@ -708,7 +707,7 @@ mod tests {
             output: chat_output.clone(),
         });
 
-        let chat_messages = dicl_config.prepare_message(&chat_example).unwrap();
+        let chat_messages = DiclConfig::prepare_message(&chat_example).unwrap();
 
         assert_eq!(chat_messages.len(), 2);
 
@@ -744,7 +743,7 @@ mod tests {
             output: json_output.clone(),
         });
 
-        let json_messages = dicl_config.prepare_message(&json_example).unwrap();
+        let json_messages = DiclConfig::prepare_message(&json_example).unwrap();
 
         // Assertions for JsonExample
         assert_eq!(json_messages.len(), 2);
@@ -769,9 +768,6 @@ mod tests {
 
     #[test]
     fn test_prepare_input_message() {
-        // Create an instance of DiclConfig (assuming default implementation is available)
-        let dicl_config = DiclConfig::default();
-
         // Mock Input data
         let input_data = ResolvedInput {
             system: Some(json!({"assistant_name": "Dr. Mehta"})),
@@ -799,7 +795,7 @@ mod tests {
         };
 
         // Call the prepare_input_message function
-        let request_message = dicl_config.prepare_input_message(&input_data).unwrap();
+        let request_message = DiclConfig::prepare_input_message(&input_data).unwrap();
 
         // The role should be User
         assert_eq!(request_message.role, Role::User);
