@@ -1,5 +1,4 @@
 use std::fmt;
-use std::path::Path;
 
 use futures::future::join_all;
 use rand::Rng;
@@ -7,7 +6,7 @@ use serde::{Deserialize, Serialize};
 use serde_json::{json, Value};
 use tokio::time::{timeout, Duration};
 
-use crate::config_parser::PathWithContents;
+use crate::config::{ErrorContext, PathWithContents, SchemaData};
 use crate::embeddings::EmbeddingModelTable;
 use crate::endpoints::inference::{InferenceClients, InferenceModels};
 use crate::error::IMPOSSIBLE_ERROR_MESSAGE;
@@ -31,15 +30,14 @@ use crate::{
     variant::chat_completion::ChatCompletionConfig,
 };
 
-use crate::config_parser::LoadableConfig;
-use crate::variant::chat_completion::{TemplateSchemaInfo, UninitializedChatCompletionConfig};
+use crate::variant::chat_completion::UninitializedChatCompletionConfig;
 
 use super::{
     infer_model_request, infer_model_request_stream, prepare_model_inference_request,
     InferModelRequestArgs, InferenceConfig, ModelUsedInfo, Variant,
 };
 
-#[derive(Debug, Serialize, Deserialize)]
+#[derive(Debug, Serialize)]
 #[cfg_attr(test, derive(ts_rs::TS))]
 #[cfg_attr(test, ts(export))]
 pub struct MixtureOfNConfig {
@@ -49,8 +47,9 @@ pub struct MixtureOfNConfig {
     pub fuser: FuserConfig,
 }
 
-#[derive(Debug, Deserialize)]
+#[derive(Clone, Debug, Deserialize, Serialize, ts_rs::TS)]
 #[serde(deny_unknown_fields)]
+#[ts(export)]
 pub struct UninitializedMixtureOfNConfig {
     #[serde(default)]
     pub weight: Option<f64>,
@@ -64,7 +63,7 @@ fn default_timeout() -> f64 {
     300.0
 }
 
-#[derive(Debug, Serialize, Deserialize)]
+#[derive(Debug, Serialize)]
 #[cfg_attr(test, derive(ts_rs::TS))]
 #[cfg_attr(test, ts(export))]
 pub struct FuserConfig {
@@ -72,21 +71,34 @@ pub struct FuserConfig {
     pub inner: ChatCompletionConfig,
 }
 
-#[derive(Debug, Deserialize)]
+#[derive(Clone, Debug, Deserialize, Serialize, ts_rs::TS)]
 #[serde(deny_unknown_fields)]
+#[ts(export)]
 pub struct UninitializedFuserConfig {
     #[serde(flatten)]
     pub inner: UninitializedChatCompletionConfig,
 }
 
-impl LoadableConfig<MixtureOfNConfig> for UninitializedMixtureOfNConfig {
-    fn load<P: AsRef<Path>>(self, base_path: P) -> Result<MixtureOfNConfig, Error> {
+impl UninitializedMixtureOfNConfig {
+    pub fn load(
+        self,
+        schemas: &SchemaData,
+        error_context: &ErrorContext,
+    ) -> Result<MixtureOfNConfig, Error> {
         Ok(MixtureOfNConfig {
             weight: self.weight,
             timeout_s: self.timeout_s,
             candidates: self.candidates,
             fuser: FuserConfig {
-                inner: self.fuser.inner.load(base_path)?,
+                inner: self.fuser.inner.load(
+                    schemas,
+                    // Our stored fuser is a plain `UninitializedChatCompletionConfig`, so we need
+                    // to explicitly add `fuser` to any error messages it produces.
+                    &ErrorContext {
+                        function_name: error_context.function_name.clone(),
+                        variant_name: format!("{}.fuser", error_context.variant_name),
+                    },
+                )?,
             },
         })
     }
@@ -98,7 +110,7 @@ impl Variant for MixtureOfNConfig {
         input: &ResolvedInput,
         models: &'request InferenceModels<'a>,
         function: &'a FunctionConfig,
-        inference_config: &'request InferenceConfig<'static, 'request>,
+        inference_config: &'request InferenceConfig<'request>,
         clients: &'request InferenceClients<'request>,
         _inference_params: InferenceParams,
     ) -> Result<InferenceResult, Error> {
@@ -131,7 +143,7 @@ impl Variant for MixtureOfNConfig {
         input: &ResolvedInput,
         models: &'request InferenceModels<'_>,
         function: &FunctionConfig,
-        inference_config: &'request InferenceConfig<'static, 'request>,
+        inference_config: &'request InferenceConfig<'request>,
         clients: &'request InferenceClients<'request>,
         inference_params: InferenceParams,
     ) -> Result<(InferenceResultStream, ModelUsedInfo), Error> {
@@ -193,7 +205,7 @@ impl Variant for MixtureOfNConfig {
                     variant_name: variant_name.to_string(),
                     message: e.to_string(),
                 })
-            })?
+            })?;
         }
         // Validate the evaluator variant
         self.fuser
@@ -222,7 +234,7 @@ impl Variant for MixtureOfNConfig {
         _input: &[ResolvedInput],
         _models: &'a InferenceModels<'a>,
         _function: &'a FunctionConfig,
-        _inference_configs: &'a [InferenceConfig<'a, 'a>],
+        _inference_configs: &'a [InferenceConfig<'a>],
         _clients: &'a InferenceClients<'a>,
         _inference_params: Vec<InferenceParams>,
     ) -> Result<StartBatchModelInferenceWithMetadata<'a>, Error> {
@@ -371,7 +383,7 @@ impl MixtureOfNConfig {
         input: &ResolvedInput,
         models: &'request InferenceModels<'a>,
         function: &'a FunctionConfig,
-        inference_config: &'request InferenceConfig<'static, 'request>,
+        inference_config: &'request InferenceConfig<'request>,
         clients: &'request InferenceClients<'request>,
     ) -> Result<Vec<InferenceResult>, Error> {
         // Get all the variants we are going to infer
@@ -400,7 +412,7 @@ impl MixtureOfNConfig {
                 // However, the 'A, C' and 'C, D' evaluations will all have distinct cache keys:
                 // (A, 2), (C, 3), (C, 2), (D, 4)
                 let mut config = inference_config.clone();
-                config.variant_name = Some(candidate);
+                config.variant_name = candidate;
                 config.extra_cache_key = Some(format!("candidate_{i}"));
                 Ok((candidate.to_string(), variant, config))
             })
@@ -439,7 +451,7 @@ impl MixtureOfNConfig {
             match result {
                 Ok(inner_result) => {
                     if let Ok(res) = inner_result {
-                        successful_results.push(res)
+                        successful_results.push(res);
                     }
                 }
                 Err(_timeout_error) => {
@@ -463,7 +475,7 @@ impl MixtureOfNConfig {
         input: &ResolvedInput,
         function: &'a FunctionConfig,
         models: &'a ModelTable,
-        inference_config: &'request InferenceConfig<'a, 'request>,
+        inference_config: &'request InferenceConfig<'request>,
         clients: &'request InferenceClients<'request>,
         mut candidates: Vec<InferenceResult>,
         stream: bool,
@@ -559,7 +571,7 @@ async fn inner_fuse_candidates<'a, 'request>(
     input: &'request ResolvedInput,
     models: &'a ModelTable,
     function: &'a FunctionConfig,
-    inference_config: &'request InferenceConfig<'a, 'request>,
+    inference_config: &'request InferenceConfig<'request>,
     clients: &'request InferenceClients<'request>,
     candidates: &[InferenceResult],
 ) -> Result<InferenceResult, Error> {
@@ -610,7 +622,7 @@ async fn inner_fuse_candidates_stream<'a, 'request>(
     input: &'request ResolvedInput,
     models: &'a ModelTable,
     function: &'a FunctionConfig,
-    inference_config: &'request InferenceConfig<'a, 'request>,
+    inference_config: &'request InferenceConfig<'request>,
     clients: &'request InferenceClients<'request>,
     candidates: &[InferenceResult],
 ) -> Result<(InferenceResultStream, ModelUsedInfo), Error> {
@@ -657,11 +669,8 @@ impl FuserConfig {
         templates: &TemplateConfig,
         system: Option<&Value>,
         max_index: usize,
-        template_schema_info: TemplateSchemaInfo,
     ) -> Result<String, Error> {
-        let inner_system_message =
-            self.inner
-                .prepare_system_message(templates, system, template_schema_info)?;
+        let inner_system_message = self.inner.prepare_system_message(templates, system)?;
         let template_context = match inner_system_message {
             Some(inner_system_message) => {
                 json!({"inner_system_message": inner_system_message, "max_index": max_index})
@@ -698,7 +707,6 @@ impl FuserConfig {
     ///
     /// Returns an `Error` if any of the candidate outputs fail to serialize or if templating fails.
     fn prepare_candidate_message(
-        &self,
         templates: &TemplateConfig,
         candidates: &[InferenceResult],
     ) -> Result<(RequestMessage, Vec<usize>), Error> {
@@ -759,7 +767,7 @@ impl FuserConfig {
         &'a self,
         input: &'request ResolvedInput,
         function: &'a FunctionConfig,
-        inference_config: &'request InferenceConfig<'a, 'request>,
+        inference_config: &'request InferenceConfig<'request>,
         candidates: &[InferenceResult],
         inference_params: &mut InferenceParams,
         stream: bool,
@@ -769,23 +777,19 @@ impl FuserConfig {
     {
         // Do this before we prepare the system message so we can use the correct max index in the system message
         let (candidate_message, included_indices) =
-            self.prepare_candidate_message(inference_config.templates, candidates)?;
+            Self::prepare_candidate_message(inference_config.templates, candidates)?;
         let max_index = included_indices.len().saturating_sub(1);
         let system = Some(self.prepare_system_message(
             inference_config.templates,
             input.system.as_ref(),
             max_index,
-            function.template_schema_info(),
         )?);
         let messages = input
             .messages
             .iter()
             .map(|message| {
-                self.inner.prepare_request_message(
-                    inference_config.templates,
-                    message,
-                    function.template_schema_info(),
-                )
+                self.inner
+                    .prepare_request_message(inference_config.templates, message)
             })
             .chain(std::iter::once(Ok(candidate_message)))
             .collect::<Result<Vec<_>, _>>()?;
@@ -818,6 +822,7 @@ impl FuserConfig {
             inference_extra_headers: inference_config
                 .extra_headers
                 .clone()
+                .into_owned()
                 .filter(inference_config.variant_name),
         };
         let model_inference_request = prepare_model_inference_request(
@@ -845,7 +850,8 @@ mod tests {
 
     use crate::{
         cache::{CacheEnabledMode, CacheOptions},
-        clickhouse::ClickHouseConnectionInfo,
+        config::SchemaData,
+        db::clickhouse::ClickHouseConnectionInfo,
         endpoints::inference::{InferenceCredentials, InferenceIds},
         function::{FunctionConfigChat, FunctionConfigJson},
         inference::types::{
@@ -853,10 +859,11 @@ mod tests {
             Latency, ModelInferenceResponseWithMetadata, Text, Thought,
         },
         jsonschema_util::StaticJSONSchema,
-        minijinja_util::tests::get_test_template_config,
+        minijinja_util::tests::{get_test_template_config, test_system_template_schema},
         model::{ModelConfig, ModelProvider, ProviderConfig},
         providers::dummy::DummyProvider,
         tool::{ToolCallConfig, ToolCallOutput, ToolChoice},
+        variant::chat_completion::{ChatTemplates, TemplateWithSchema},
     };
 
     use super::*;
@@ -865,12 +872,6 @@ mod tests {
     fn test_prepare_system_message() {
         let templates = get_test_template_config();
 
-        let all_schemas = TemplateSchemaInfo {
-            has_system_schema: true,
-            has_user_schema: true,
-            has_assistant_schema: true,
-        };
-
         // Test without templates, string message
         let fuser_config = FuserConfig {
             inner: ChatCompletionConfig {
@@ -881,12 +882,8 @@ mod tests {
         };
         let input_message = Value::String("You are a helpful assistant.".to_string());
         let max_index = 2;
-        let result = fuser_config.prepare_system_message(
-            &templates,
-            Some(&input_message),
-            max_index,
-            all_schemas,
-        );
+        let result =
+            fuser_config.prepare_system_message(&templates, Some(&input_message), max_index);
         let prepared_message = result.unwrap();
         let expected_message = templates
             .template_message(
@@ -906,12 +903,8 @@ mod tests {
         };
         let input_message = json!({"message": "You are a helpful assistant."});
         let max_index = 3;
-        let result = fuser_config.prepare_system_message(
-            &templates,
-            Some(&input_message),
-            max_index,
-            all_schemas,
-        );
+        let result =
+            fuser_config.prepare_system_message(&templates, Some(&input_message), max_index);
         assert!(result.is_err());
         let prepared_message = result.unwrap_err();
         assert_eq!(
@@ -928,7 +921,7 @@ mod tests {
             },
         };
         let max_index = 5;
-        let result = fuser_config.prepare_system_message(&templates, None, max_index, all_schemas);
+        let result = fuser_config.prepare_system_message(&templates, None, max_index);
         let expected_message = templates
             .template_message(
                 "t0:mixture_of_n_fuser_system",
@@ -946,24 +939,25 @@ mod tests {
             inner: ChatCompletionConfig {
                 model: "dummy".into(),
                 weight: Some(1.0),
-                system_template: Some(PathWithContents {
-                    path: system_template_name.into(),
-                    contents: String::new(),
-                }),
+                templates: ChatTemplates {
+                    system: Some(TemplateWithSchema {
+                        template: PathWithContents {
+                            path: system_template_name.into(),
+                            contents: String::new(),
+                        },
+                        schema: Some(test_system_template_schema()),
+                    }),
+                    ..Default::default()
+                },
                 ..Default::default()
             },
         };
 
         let max_index = 6;
         let input_message = serde_json::json!({"assistant_name": "ChatGPT"});
-        let result = fuser_config.prepare_system_message(
-            &templates,
-            Some(&input_message),
-            max_index,
-            all_schemas,
-        );
-        assert!(result.is_ok());
-        let prepared_message = result.unwrap();
+        let prepared_message = fuser_config
+            .prepare_system_message(&templates, Some(&input_message), max_index)
+            .unwrap();
         let inner_system_message = templates
             .template_message(
                 system_template_name,
@@ -985,16 +979,22 @@ mod tests {
             inner: ChatCompletionConfig {
                 model: "dummy".into(),
                 weight: Some(1.0),
-                system_template: Some(PathWithContents {
-                    path: system_template_name.into(),
-                    contents: String::new(),
-                }),
+                templates: ChatTemplates {
+                    system: Some(TemplateWithSchema {
+                        template: PathWithContents {
+                            path: system_template_name.into(),
+                            contents: String::new(),
+                        },
+                        schema: None,
+                    }),
+                    ..Default::default()
+                },
                 ..Default::default()
             },
         };
 
         let max_index = 10;
-        let result = fuser_config.prepare_system_message(&templates, None, max_index, all_schemas);
+        let result = fuser_config.prepare_system_message(&templates, None, max_index);
         assert!(result.is_ok());
         let prepared_message = result.unwrap();
         let inner_system_message = templates
@@ -1013,15 +1013,6 @@ mod tests {
     async fn test_prepare_candidate_message() {
         let templates = get_test_template_config();
 
-        // Create an FuserConfig
-        let fuser_config = FuserConfig {
-            inner: ChatCompletionConfig {
-                model: "dummy".into(),
-                weight: Some(1.0),
-                ..Default::default()
-            },
-        };
-
         // Prepare some candidate InferenceResults
         let model_inference_response = ModelInferenceResponseWithMetadata {
             id: Uuid::now_v7(),
@@ -1092,7 +1083,7 @@ mod tests {
         let candidates = vec![candidate1, candidate2];
 
         // Call prepare_candidate_message
-        let result = fuser_config.prepare_candidate_message(&templates, &candidates);
+        let result = FuserConfig::prepare_candidate_message(&templates, &candidates);
         assert!(result.is_ok());
         let (request_message, included_indices) = result.unwrap();
         assert_eq!(included_indices, vec![0, 1]);
@@ -1107,15 +1098,6 @@ mod tests {
     async fn test_prepare_candidate_message_json() {
         let templates = get_test_template_config();
 
-        // Create a FuserConfig
-        let fuser_config = FuserConfig {
-            inner: ChatCompletionConfig {
-                model: "dummy_json".into(),
-                weight: Some(1.0),
-                ..Default::default()
-            },
-        };
-
         // Prepare some candidate InferenceResults - some valid, some malformed
         let model_inference_response_valid = ModelInferenceResponseWithMetadata {
             id: Uuid::now_v7(),
@@ -1188,7 +1170,7 @@ mod tests {
         let candidates = vec![candidate1, candidate2];
 
         // Call prepare_candidate_message
-        let result = fuser_config.prepare_candidate_message(&templates, &candidates);
+        let result = FuserConfig::prepare_candidate_message(&templates, &candidates);
         assert!(result.is_ok());
         let (request_message, included_indices) = result.unwrap();
 
@@ -1221,10 +1203,8 @@ mod tests {
         let templates = get_test_template_config();
         let json_function_config = FunctionConfig::Json(FunctionConfigJson {
             variants: HashMap::new(),
-            system_schema: None,
-            user_schema: None,
-            assistant_schema: None,
-            output_schema: StaticJSONSchema::from_value(&json!({})).unwrap(),
+            schemas: SchemaData::default(),
+            output_schema: StaticJSONSchema::from_value(json!({})).unwrap(),
             implicit_tool_call_config: ToolCallConfig::default(),
             description: None,
         });
@@ -1342,7 +1322,7 @@ mod tests {
             tool_config: None,
             dynamic_output_schema: None,
             function_name: "",
-            variant_name: Some(""),
+            variant_name: "",
             extra_body: Default::default(),
             extra_headers: Default::default(),
             extra_cache_key: None,
@@ -1502,9 +1482,7 @@ mod tests {
         };
         let chat_function_config = FunctionConfig::Chat(FunctionConfigChat {
             variants: HashMap::new(),
-            system_schema: None,
-            user_schema: None,
-            assistant_schema: None,
+            schemas: SchemaData::default(),
             tools: vec![],
             tool_choice: ToolChoice::None,
             parallel_tool_calls: None,
