@@ -3,8 +3,8 @@ use std::collections::HashMap;
 use std::sync::Arc;
 
 use serde_json::json;
-use tensorzero_core::clickhouse::ClickHouseConnectionInfo;
-use tensorzero_core::config_parser::Config;
+use tensorzero_core::config::Config;
+use tensorzero_core::db::clickhouse::{ClickHouseConnectionInfo, TableName};
 /// End-to-end tests for particular internal functionality in the batch inference endpoint
 /// These are not tests of the public API (those should go in tests/e2e/providers/batch.rs)
 use tensorzero_core::endpoints::batch_inference::{
@@ -25,7 +25,7 @@ use tensorzero_core::jsonschema_util::StaticJSONSchema;
 use tokio::time::{sleep, Duration};
 use uuid::Uuid;
 
-use tensorzero_core::clickhouse::test_helpers::{
+use tensorzero_core::db::clickhouse::test_helpers::{
     get_clickhouse, select_chat_inference_clickhouse, select_json_inference_clickhouse,
     select_model_inferences_clickhouse,
 };
@@ -103,7 +103,7 @@ async fn test_get_batch_request() {
     };
     let rows = vec![row];
     clickhouse
-        .write(&rows, "BatchModelInference")
+        .write_batched(&rows, TableName::BatchModelInference)
         .await
         .unwrap();
     // Sleep a bit to ensure the write has propagated
@@ -261,7 +261,7 @@ async fn write_2_batch_model_inference_rows(
     };
     let rows = vec![row1, row2];
     clickhouse
-        .write(&rows, "BatchModelInference")
+        .write_batched(&rows, TableName::BatchModelInference)
         .await
         .unwrap();
     rows
@@ -382,7 +382,7 @@ async fn test_write_read_completed_batch_inference_chat() {
             );
             match &chat_inference_response.content[0] {
                 ContentBlockChatOutput::Text(text_block) => {
-                    assert_eq!(text_block.text, "hello world")
+                    assert_eq!(text_block.text, "hello world");
                 }
                 _ => panic!("Unexpected content block type"),
             }
@@ -397,7 +397,7 @@ async fn test_write_read_completed_batch_inference_chat() {
             assert_eq!(chat_inference_response.inference_id, inference_id2);
             match &chat_inference_response.content[0] {
                 ContentBlockChatOutput::Text(text_block) => {
-                    assert_eq!(text_block.text, "goodbye world")
+                    assert_eq!(text_block.text, "goodbye world");
                 }
                 _ => panic!("Unexpected content block type"),
             }
@@ -515,7 +515,7 @@ async fn test_write_read_completed_batch_inference_json() {
         status,
         errors: vec![],
     });
-    let output_schema = StaticJSONSchema::from_value(&json!({
+    let output_schema = StaticJSONSchema::from_value(json!({
         "type": "object",
         "properties": {
             "answer": {
