@@ -3,32 +3,43 @@
 use std::cell::Cell;
 use std::future::Future;
 use std::sync::Arc;
+use std::time::Duration;
 
+use async_trait::async_trait;
 use paste::paste;
 use secrecy::{ExposeSecret, SecretString};
 use serde_json::json;
-use tensorzero_core::clickhouse::migration_manager::migration_trait::Migration;
+use tensorzero_core::db::clickhouse::migration_manager::migration_trait::Migration;
+use tensorzero_core::db::clickhouse::migration_manager::{
+    RunMigrationArgs, RunMigrationManagerArgs,
+};
 use tensorzero_core::endpoints::status::TENSORZERO_VERSION;
+use tensorzero_core::error::Error;
 use tokio::runtime::Handle;
+use tokio::time::sleep;
 use tracing_test::traced_test;
 use uuid::Uuid;
 
-use tensorzero_core::clickhouse::migration_manager::migrations::migration_0000::Migration0000;
-use tensorzero_core::clickhouse::migration_manager::migrations::migration_0002::Migration0002;
-use tensorzero_core::clickhouse::migration_manager::migrations::migration_0003::Migration0003;
-use tensorzero_core::clickhouse::migration_manager::migrations::migration_0004::Migration0004;
-use tensorzero_core::clickhouse::migration_manager::migrations::migration_0005::Migration0005;
-use tensorzero_core::clickhouse::migration_manager::migrations::migration_0006::Migration0006;
-use tensorzero_core::clickhouse::migration_manager::migrations::migration_0008::Migration0008;
-use tensorzero_core::clickhouse::migration_manager::migrations::migration_0009::Migration0009;
-use tensorzero_core::clickhouse::migration_manager::migrations::migration_0011::Migration0011;
-use tensorzero_core::clickhouse::migration_manager::migrations::migration_0013::Migration0013;
+use tensorzero_core::db::clickhouse::migration_manager::migrations::check_table_exists;
+use tensorzero_core::db::clickhouse::migration_manager::migrations::migration_0000::Migration0000;
+use tensorzero_core::db::clickhouse::migration_manager::migrations::migration_0002::Migration0002;
+use tensorzero_core::db::clickhouse::migration_manager::migrations::migration_0003::Migration0003;
+use tensorzero_core::db::clickhouse::migration_manager::migrations::migration_0004::Migration0004;
+use tensorzero_core::db::clickhouse::migration_manager::migrations::migration_0005::Migration0005;
+use tensorzero_core::db::clickhouse::migration_manager::migrations::migration_0006::Migration0006;
+use tensorzero_core::db::clickhouse::migration_manager::migrations::migration_0008::Migration0008;
+use tensorzero_core::db::clickhouse::migration_manager::migrations::migration_0009::Migration0009;
+use tensorzero_core::db::clickhouse::migration_manager::migrations::migration_0011::Migration0011;
+use tensorzero_core::db::clickhouse::migration_manager::migrations::migration_0013::Migration0013;
+use tensorzero_core::inference::types::ModelInferenceDatabaseInsert;
 
-use tensorzero_core::clickhouse::migration_manager::{
-    self, make_all_migrations, MigrationRecordDatabaseInsert,
+use tensorzero_core::db::clickhouse::migration_manager::{
+    self, get_all_migration_records, make_all_migrations, MigrationRecordDatabaseInsert,
+};
+use tensorzero_core::db::clickhouse::test_helpers::{get_clickhouse, CLICKHOUSE_URL};
+use tensorzero_core::db::clickhouse::{
+    make_clickhouse_http_client, ClickHouseConnectionInfo, Rows, TableName,
 };
-use tensorzero_core::clickhouse::test_helpers::{get_clickhouse, CLICKHOUSE_URL};
-use tensorzero_core::clickhouse::{make_clickhouse_http_client, ClickHouseConnectionInfo};
 
 pub struct DeleteDbOnDrop {
     database: String,
@@ -58,7 +69,7 @@ impl Drop for DeleteDbOnDrop {
                         .unwrap();
                 }
                 eprintln!("Database dropped: {database}");
-            })
+            });
         });
     }
 }
@@ -77,11 +88,14 @@ pub fn get_clean_clickhouse(allow_db_missing: bool) -> (ClickHouseConnectionInfo
     let mut clickhouse_url = url::Url::parse(&CLICKHOUSE_URL).unwrap();
     clickhouse_url.set_path("");
     clickhouse_url.set_query(Some(format!("database={database}").as_str()));
+    let cluster_name = std::env::var("TENSORZERO_CLICKHOUSE_CLUSTER_NAME").ok();
 
     let clickhouse = ClickHouseConnectionInfo::Production {
         database_url: SecretString::from(clickhouse_url.to_string()),
         database: database.clone(),
+        cluster_name,
         client: make_clickhouse_http_client().unwrap(),
+        batch_sender: None,
     };
     (
         clickhouse.clone(),
@@ -98,7 +112,7 @@ macro_rules! invoke_all_separate_tests {
         // For each value in the literal array, generate a new `#[traced_test]` function
         // that calls the target function with that value, along with the `logs_contain`
         // helper generated by the `traced_test` macro.
-        const _MIGRATIONS_NUM_ARRAY: [usize; tensorzero_core::clickhouse::migration_manager::NUM_MIGRATIONS] = [$($migration_num),*];
+        const _MIGRATIONS_NUM_ARRAY: [usize; tensorzero_core::db::clickhouse::migration_manager::NUM_MIGRATIONS] = [$($migration_num),*];
         $(
             paste! {
                 #[tokio::test(flavor = "multi_thread")]
@@ -166,7 +180,9 @@ async fn insert_large_fixtures(clickhouse: &ClickHouseConnectionInfo) {
     let ClickHouseConnectionInfo::Production {
         database_url,
         database,
+        cluster_name: _,
         client: _,
+        batch_sender: _,
     } = clickhouse
     else {
         panic!("ClickHouseConnectionInfo is not a Production connection");
@@ -186,6 +202,8 @@ async fn insert_large_fixtures(clickhouse: &ClickHouseConnectionInfo) {
     let insert_futures = [
         ("large_chat_inference_v2.parquet", "ChatInference"),
         ("large_json_inference_v2.parquet", "JsonInference"),
+        ("large_chat_model_inference_v2.parquet", "ModelInference"),
+        ("large_json_model_inference_v2.parquet", "ModelInference"),
         (
             "large_chat_boolean_feedback.parquet",
             "BooleanMetricFeedback",
@@ -229,9 +247,9 @@ async fn insert_large_fixtures(clickhouse: &ClickHouseConnectionInfo) {
                 database,
                 "--query",
                 &format!(
-                    r#"
+                    r"
         INSERT INTO {table} FROM INFILE '/s3-fixtures/{file}' FORMAT Parquet
-    "#
+    "
                 ),
             ]);
             assert!(
@@ -355,7 +373,7 @@ async fn run_migration_0020_with_data<R: Future<Output = bool>, F: FnOnce() -> R
 
     let matching_chat_by_id = clickhouse
         .run_query_synchronous_no_params(
-            format!("SELECT id_uint, toUInt128(episode_id) as episode_id_uint FROM InferenceById FINAL WHERE function_type = 'chat' AND id_uint = '{sample_chat_id}' LIMIT 1 FORMAT JSONEachRow"),
+            format!("SELECT id_uint, toUInt128(episode_id) as episode_id_uint FROM InferenceById WHERE function_type = 'chat' AND id_uint = '{sample_chat_id}' LIMIT 1 FORMAT JSONEachRow"),
         )
         .await
         .unwrap();
@@ -373,7 +391,7 @@ async fn run_migration_0020_with_data<R: Future<Output = bool>, F: FnOnce() -> R
 
     let matching_chat_by_episode_id = clickhouse
         .run_query_synchronous_no_params(
-            format!("SELECT * FROM InferenceByEpisodeId FINAL WHERE function_type = 'chat' AND episode_id_uint = '{sample_chat_episode_id}' LIMIT 1 FORMAT JSONEachRow"),
+            format!("SELECT * FROM InferenceByEpisodeId WHERE function_type = 'chat' AND episode_id_uint = '{sample_chat_episode_id}' LIMIT 1 FORMAT JSONEachRow"),
         )
         .await
         .unwrap();
@@ -401,7 +419,7 @@ async fn run_migration_0020_with_data<R: Future<Output = bool>, F: FnOnce() -> R
 
     let matching_json_by_id = clickhouse
         .run_query_synchronous_no_params(
-            format!("SELECT id_uint, toUInt128(episode_id) as episode_id_uint FROM InferenceById FINAL WHERE function_type = 'json' AND id_uint = '{sample_json_id}' LIMIT 1 FORMAT JSONEachRow"),
+            format!("SELECT id_uint, toUInt128(episode_id) as episode_id_uint FROM InferenceById WHERE function_type = 'json' AND id_uint = '{sample_json_id}' LIMIT 1 FORMAT JSONEachRow"),
         )
         .await
         .unwrap();
@@ -417,7 +435,7 @@ async fn run_migration_0020_with_data<R: Future<Output = bool>, F: FnOnce() -> R
 
     let matching_json_by_episode_id = clickhouse
         .run_query_synchronous_no_params(
-            format!("SELECT * FROM InferenceByEpisodeId FINAL WHERE function_type = 'json' AND episode_id_uint = '{sample_json_episode_id}' LIMIT 1 FORMAT JSONEachRow"),
+            format!("SELECT * FROM InferenceByEpisodeId WHERE function_type = 'json' AND episode_id_uint = '{sample_json_episode_id}' LIMIT 1 FORMAT JSONEachRow"),
         )
         .await
         .unwrap();
@@ -452,7 +470,10 @@ async fn run_rollback_instructions(
 }
 async fn test_rollback_helper(migration_num: usize, logs_contain: fn(&str) -> bool) {
     let (fresh_clickhouse, _cleanup_fresh_clickhouse) = get_clean_clickhouse(true);
-    fresh_clickhouse.create_database().await.unwrap();
+    fresh_clickhouse
+        .create_database_and_migrations_table()
+        .await
+        .unwrap();
     let migrations = make_all_migrations(&fresh_clickhouse);
     println!(
         "Running migrations up to {}",
@@ -461,9 +482,15 @@ async fn test_rollback_helper(migration_num: usize, logs_contain: fn(&str) -> bo
     for migration in &migrations[..=migration_num] {
         let name = migration.name();
         println!("Running migration: {name}");
-        migration_manager::run_migration(&fresh_clickhouse, migration.as_ref(), false)
-            .await
-            .unwrap();
+        migration_manager::run_migration(RunMigrationArgs {
+            clickhouse: &fresh_clickhouse,
+            migration: migration.as_ref(),
+            clean_start: false,
+            manual_run: false,
+            is_replicated: false,
+        })
+        .await
+        .unwrap();
         // Migration0029 only runs if `StaticEvaluationHumanFeedbackFloatView` or `StaticEvaluationHumanFeedbackBooleanView`
         // exists, which were created by the banned migration Migration0023
         let should_succeed = migration.name() != "Migration0029";
@@ -481,7 +508,10 @@ async fn test_rollback_helper(migration_num: usize, logs_contain: fn(&str) -> bo
     // to try to run commands on a non-existent database.
     // We make sure that the database exists at the end of the rollback to prevent '_cleanup_fresh_clickhouse' from
     // panicking on drop.
-    fresh_clickhouse.create_database().await.unwrap();
+    fresh_clickhouse
+        .create_database_and_migrations_table()
+        .await
+        .unwrap();
 }
 
 // Generate tests named 'test_rollback_up_to_migration_index_n' for each migration index `n``
@@ -492,7 +522,7 @@ invoke_all_separate_tests!(
     test_rollback_up_to_migration_index_,
     [
         0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24,
-        25, 26
+        25, 26, 27, 28
     ]
 );
 
@@ -500,14 +530,23 @@ invoke_all_separate_tests!(
 #[traced_test]
 async fn test_rollback_apply_rollback() {
     let (clickhouse, _cleanup_db) = get_clean_clickhouse(false);
-    clickhouse.create_database().await.unwrap();
+    clickhouse
+        .create_database_and_migrations_table()
+        .await
+        .unwrap();
     let migrations = make_all_migrations(&clickhouse);
     for migration in migrations {
         let name = migration.name();
         println!("Running migration: {name}");
-        migration_manager::run_migration(&clickhouse, migration.as_ref(), false)
-            .await
-            .unwrap();
+        migration_manager::run_migration(RunMigrationArgs {
+            clickhouse: &clickhouse,
+            migration: migration.as_ref(),
+            clean_start: false,
+            manual_run: false,
+            is_replicated: false,
+        })
+        .await
+        .unwrap();
 
         // Migration0029 only runs if `StaticEvaluationHumanFeedbackFloatView` or `StaticEvaluationHumanFeedbackBooleanView`
         // exists, which were created by the banned migration Migration0023
@@ -523,13 +562,23 @@ async fn test_rollback_apply_rollback() {
 
         // This migration drops the entire database during rollback, so we need to re-create it
         if migration.name() == "Migration0000" {
-            clickhouse.create_database().await.unwrap();
+            sleep(Duration::from_millis(500)).await;
+            clickhouse
+                .create_database_and_migrations_table()
+                .await
+                .unwrap();
         }
 
         println!("Re-apply migration: {name}");
-        migration_manager::run_migration(&clickhouse, migration.as_ref(), false)
-            .await
-            .unwrap();
+        migration_manager::run_migration(RunMigrationArgs {
+            clickhouse: &clickhouse,
+            migration: migration.as_ref(),
+            clean_start: false,
+            manual_run: false,
+            is_replicated: false,
+        })
+        .await
+        .unwrap();
         if should_succeed {
             assert!(
                 logs_contain(&format!("Migration succeeded: {name}")),
@@ -542,11 +591,18 @@ async fn test_rollback_apply_rollback() {
 #[tokio::test(flavor = "multi_thread")]
 async fn test_clickhouse_migration_manager() {
     let (clickhouse, _cleanup_db) = get_clean_clickhouse(false);
-    clickhouse.create_database().await.unwrap();
+    clickhouse
+        .create_database_and_migrations_table()
+        .await
+        .unwrap();
     // Run it twice to test that it is a no-op the second time
-    clickhouse.create_database().await.unwrap();
+    clickhouse
+        .create_database_and_migrations_table()
+        .await
+        .unwrap();
     let migrations = make_all_migrations(&clickhouse);
     let initial_clean_start = Cell::new(true);
+    let manual_run = clickhouse.is_cluster_configured();
     // This runs all migrations up to and including the given migration number,
     // verifying that only the most recent migration is actually applied.
     let run_migrations_up_to = |migration_num: usize, logs_contain: fn(&str) -> bool| {
@@ -556,31 +612,43 @@ async fn test_clickhouse_migration_manager() {
         async move {
             // All of the previous migrations should have already been run
             for (i, migration) in migrations.iter().enumerate().take(migration_num) {
-                let clean_start =
-                    migration_manager::run_migration(clickhouse, migration.as_ref(), false)
-                        .await
-                        .unwrap();
+                let clean_start = migration_manager::run_migration(RunMigrationArgs {
+                    clickhouse,
+                    migration: migration.as_ref(),
+                    clean_start: false,
+                    manual_run,
+                    is_replicated: false,
+                })
+                .await
+                .unwrap();
                 if i == 0 {
                     // We know that the first migration was run in a previous test, so clean start should be false
                     assert!(!clean_start);
                 }
                 let name = migrations[i].name();
-                assert!(
-                    !logs_contain(&format!("Applying migration: {name}")),
-                    "Migration {name} should not have been applied"
-                );
-                assert!(
-                    !logs_contain(&format!("Migration succeeded: {name}")),
-                    "Migration {name} should not have succeeded (because it wasn't applied)"
-                );
+                // Migration0029 always runs (since we want it to write a migration row on a clean start)
+                if migration.name() != "Migration0029" {
+                    assert!(
+                        !logs_contain(&format!("Applying migration: {name}")),
+                        "Migration {name} should not have been applied"
+                    );
+                    assert!(
+                        !logs_contain(&format!("Migration succeeded: {name}")),
+                        "Migration {name} should not have succeeded (because it wasn't applied)"
+                    );
+                }
             }
+            assert!(!logs_contain("Materialized view `CumulativeUsageView` was not written because it was recently created"),
+                "CumulativeUsage backfilling failed.");
 
             let run_migration = || async {
-                migration_manager::run_migration(
+                migration_manager::run_migration(RunMigrationArgs {
                     clickhouse,
-                    migrations[migration_num].as_ref(),
-                    initial_clean_start.get(),
-                )
+                    migration: migrations[migration_num].as_ref(),
+                    clean_start: initial_clean_start.get(),
+                    is_replicated: false,
+                    manual_run,
+                })
                 .await
                 .unwrap()
             };
@@ -637,17 +705,31 @@ async fn test_clickhouse_migration_manager() {
     ) {
         // Now, run all of the migrations, and verify that none of them apply
         for (i, migration) in migrations.iter().enumerate() {
-            let clean_start =
-                migration_manager::run_migration(clickhouse, migration.as_ref(), true)
-                    .await
-                    .unwrap();
+            let clean_start = migration_manager::run_migration(RunMigrationArgs {
+                clickhouse,
+                migration: migration.as_ref(),
+                clean_start: true,
+                is_replicated: false,
+                manual_run: true,
+            })
+            .await
+            .unwrap();
             if i == 0 {
                 // We know that the first migration was run in a previous test, so clean start should be false
                 assert!(!clean_start);
             }
             let name = migrations[i].name();
-            assert!(!logs_contain(&format!("Applying migration: {name}")));
-            assert!(!logs_contain(&format!("Migration succeeded: {name}")));
+            // Migration0029 always runs
+            if name != "Migration0029" {
+                assert!(
+                    !logs_contain(&format!("Applying migration: {name}")),
+                    "Missing log for {name}"
+                );
+                assert!(
+                    !logs_contain(&format!("Migration succeeded: {name}")),
+                    "Missing success for {name}"
+                );
+            }
         }
 
         assert!(!logs_contain("Failed to apply migration"));
@@ -665,17 +747,12 @@ async fn test_clickhouse_migration_manager() {
         // for each element in the array.
         [
             0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23,
-            24, 25, 26
+            24, 25, 26, 27, 28
         ]
     );
-    let rows = get_all_migration_records(&clickhouse).await;
+    let rows = get_all_migration_records(&clickhouse).await.unwrap();
     println!("Rows: {rows:#?}");
-    // Migration0029 is currently skipped, because 'StaticEvaluationHumanFeedbackFloatView' is created
-    // by a banned migration
-    let expected_migrations = migrations
-        .iter()
-        .filter(|m| m.name() != "Migration0029")
-        .collect::<Vec<_>>();
+    let expected_migrations = &migrations;
 
     // Check that we wrote out migration records for all migrations
     assert_eq!(rows.len(), expected_migrations.len());
@@ -699,29 +776,81 @@ async fn test_clickhouse_migration_manager() {
         assert!(applied_at.is_some());
     }
     run_all(&clickhouse, &migrations).await;
-    let new_rows = get_all_migration_records(&clickhouse).await;
-    // Since we've already ran all of the migrations, we shouldn't have written any new records
+    let mut new_rows = get_all_migration_records(&clickhouse).await.unwrap();
 
-    assert_eq!(new_rows, rows);
-}
-
-async fn get_all_migration_records(
-    clickhouse: &ClickHouseConnectionInfo,
-) -> Vec<MigrationRecordDatabaseInsert> {
-    let mut rows = Vec::new();
-    for row in clickhouse
+    let response = clickhouse
         .run_query_synchronous_no_params(
-            "SELECT * FROM TensorZeroMigration ORDER BY migration_id FORMAT JSONEachRow"
-                .to_string(),
+            "SELECT count FROM CumulativeUsage FINAL WHERE type='input_tokens'".to_string(),
         )
         .await
-        .unwrap()
-        .response
-        .lines()
-    {
-        rows.push(serde_json::from_str::<MigrationRecordDatabaseInsert>(row).unwrap());
-    }
-    rows
+        .unwrap();
+    let input_token_total: u64 = response.response.trim().parse().unwrap();
+    assert_eq!(input_token_total, 200000000);
+    let response = clickhouse
+        .run_query_synchronous_no_params(
+            "SELECT count FROM CumulativeUsage FINAL WHERE type='output_tokens'".to_string(),
+        )
+        .await
+        .unwrap();
+    let output_token_total: u64 = response.response.trim().parse().unwrap();
+    assert_eq!(output_token_total, 200000000);
+    // Let's add a ModelInference row with null output tokens only then check the input tokens are correct
+    let row = ModelInferenceDatabaseInsert {
+        id: Uuid::now_v7(),
+        inference_id: Uuid::now_v7(),
+        raw_request: String::new(),
+        raw_response: String::new(),
+        system: None,
+        input_messages: String::new(),
+        output: String::new(),
+        input_tokens: Some(123),
+        output_tokens: None,
+        response_time_ms: None,
+        model_name: String::new(),
+        model_provider_name: String::new(),
+        ttft_ms: None,
+        cached: false,
+        finish_reason: None,
+    };
+    clickhouse
+        .write_non_batched(Rows::Unserialized(&[row]), TableName::ModelInference)
+        .await
+        .unwrap();
+    tokio::time::sleep(Duration::from_millis(500)).await;
+    let response = clickhouse
+        .run_query_synchronous_no_params(
+            "SELECT count FROM CumulativeUsage FINAL WHERE type='input_tokens'".to_string(),
+        )
+        .await
+        .unwrap();
+    let input_token_total: u64 = response.response.trim().parse().unwrap();
+    assert_eq!(input_token_total, 200000123);
+    let response = clickhouse
+        .run_query_synchronous_no_params(
+            "SELECT count FROM CumulativeUsage FINAL WHERE type='output_tokens'".to_string(),
+        )
+        .await
+        .unwrap();
+    let output_token_total: u64 = response.response.trim().parse().unwrap();
+    assert_eq!(output_token_total, 200000000);
+
+    // Since we've already ran all of the migrations, we shouldn't have written any new records
+    // except for Migration0029 (which runs every time)
+
+    let mut rows = rows;
+    let _new_migration0029 = new_rows
+        .iter()
+        .find(|row| row.migration_name == "Migration0029")
+        .unwrap();
+    new_rows.retain(|row| row.migration_name != "Migration0029");
+
+    let _old_migration0029 = rows
+        .iter()
+        .find(|row| row.migration_name == "Migration0029")
+        .unwrap();
+    rows.retain(|row| row.migration_name != "Migration0029");
+
+    assert_eq!(new_rows, rows);
 }
 
 #[tokio::test]
@@ -731,7 +860,7 @@ async fn test_bad_clickhouse_write() {
     let payload =
         json!({"target_id": Uuid::now_v7(), "value": true, "name": "test", "id": Uuid::now_v7()});
     let err = clickhouse
-        .write(&[payload], "BooleanMetricFeedback")
+        .write_batched(&[payload], TableName::BooleanMetricFeedback)
         .await
         .unwrap_err();
     assert!(
@@ -744,21 +873,109 @@ async fn test_bad_clickhouse_write() {
 #[tokio::test(flavor = "multi_thread")]
 async fn test_clean_clickhouse_start() {
     let (clickhouse, _cleanup_db) = get_clean_clickhouse(false);
-    migration_manager::run(&clickhouse).await.unwrap();
+    let database = clickhouse.database();
+    let is_manual = clickhouse.is_cluster_configured();
+    migration_manager::run(RunMigrationManagerArgs {
+        clickhouse: &clickhouse,
+        manual_run: is_manual,
+        skip_completed_migrations: false,
+    })
+    .await
+    .unwrap();
+
+    // We also verify here that all tables are either replicated or not replicated as expected
+    let response = clickhouse
+        .run_query_synchronous_no_params("SHOW TABLES".to_string())
+        .await
+        .unwrap();
+    let tables = response.response.split('\n');
+    for table in tables {
+        let table = table.trim();
+        if table.is_empty() {
+            continue;
+        }
+        let create_table_info = clickhouse
+            .run_query_synchronous_no_params(format!("SHOW CREATE TABLE {table}"))
+            .await
+            .unwrap()
+            .response;
+        println!("create_table_info: {create_table_info}");
+        // We only need to worry about MergeTree tables when checking replication
+        if !create_table_info.contains("MergeTree") {
+            continue;
+        }
+        let engine_is_replicated = create_table_info.contains("Replicated");
+        let replica_info = clickhouse.run_query_synchronous_no_params(format!(
+            "SELECT total_replicas FROM system.replicas WHERE database = '{database}' AND table = '{table}'"
+        )).await.unwrap();
+        if clickhouse.is_cluster_configured() {
+            assert!(
+                engine_is_replicated,
+                "Table {table} is not replicated but ClickHouse is configured for replication."
+            );
+            let replica_count: u8 = replica_info.response.trim().parse().unwrap();
+            assert_eq!(replica_count, 2);
+        }
+    }
+}
+
+#[tokio::test(flavor = "multi_thread")]
+async fn test_startup_without_migration_table() {
+    let (clickhouse, _cleanup_db) = get_clean_clickhouse(false);
+    let is_manual = clickhouse.is_cluster_configured();
+    // Run the migrations so we can get the database into a "dirty" state
+    migration_manager::run(RunMigrationManagerArgs {
+        clickhouse: &clickhouse,
+        manual_run: is_manual,
+        skip_completed_migrations: false,
+    })
+    .await
+    .unwrap();
+
+    // Drop the TensorZeroMigration table
+    clickhouse
+        .run_query_synchronous_no_params(format!(
+            "DROP TABLE TensorZeroMigration {} SYNC",
+            clickhouse.get_on_cluster_name()
+        ))
+        .await
+        .unwrap();
+
+    // Run the migrations again to ensure that they don't panic and that the table is recreated
+    migration_manager::run(RunMigrationManagerArgs {
+        clickhouse: &clickhouse,
+        manual_run: is_manual,
+        skip_completed_migrations: false,
+    })
+    .await
+    .unwrap();
+
+    // Make sure the table exists
+    assert!(
+        check_table_exists(&clickhouse, "TensorZeroMigration", "TEST")
+            .await
+            .unwrap()
+    );
 }
 
 #[tokio::test(flavor = "multi_thread")]
 async fn test_deployment_id_oldest() {
     let (clickhouse, _cleanup_db) = get_clean_clickhouse(false);
-    migration_manager::run(&clickhouse).await.unwrap();
+    migration_manager::run(RunMigrationManagerArgs {
+        clickhouse: &clickhouse,
+        manual_run: true,
+        skip_completed_migrations: false,
+    })
+    .await
+    .unwrap();
     // Add a row to the DeploymentID table and make sure that it isn't returned
     let new_deployment_id = "foo";
     clickhouse
-        .write(
-            &[serde_json::json!({
+        .write_non_batched(
+            Rows::Unserialized(&[serde_json::json!({
                 "deployment_id": new_deployment_id,
-            })],
-            "DeploymentID",
+            })]),
+            TableName::DeploymentID,
         )
         .await
         .unwrap();
@@ -777,6 +994,10 @@ async fn test_deployment_id_oldest() {
 
 #[tokio::test(flavor = "multi_thread")]
 async fn test_concurrent_clickhouse_migrations() {
+    if std::env::var("TENSORZERO_CLICKHOUSE_CLUSTER_NAME").is_ok() {
+        // We can't run concurrent migrations on a cluster.
+        return;
+    }
     let (clickhouse, _cleanup_db) = get_clean_clickhouse(false);
     let clickhouse = Arc::new(clickhouse);
     let num_concurrent_starts = 50;
@@ -784,12 +1005,41 @@ async fn test_concurrent_clickhouse_migrations() {
     for _ in 0..num_concurrent_starts {
         let clickhouse_clone = clickhouse.clone();
         handles.push(tokio::spawn(async move {
-            migration_manager::run(&clickhouse_clone).await.unwrap();
+            migration_manager::run(RunMigrationManagerArgs {
+                clickhouse: &clickhouse_clone,
+                manual_run: false,
+                skip_completed_migrations: false,
+            })
+            .await
+            .unwrap();
         }));
     }
     for handle in handles {
         handle.await.unwrap();
     }
+
+    // We should have written at least one duplicate migration record to `TensorZeroMigration`
+    // due to multiple copies of the same migration running concurrently.
+    let total_runs = clickhouse
+        .run_query_synchronous_no_params("SELECT COUNT(*) FROM TensorZeroMigration".to_string())
+        .await
+        .unwrap()
+        .response
+        .trim()
+        .parse::<u64>()
+        .unwrap();
+
+    let all_migrations = migration_manager::get_all_migration_records(&clickhouse)
+        .await
+        .unwrap();
+    assert!(
+        total_runs as usize > all_migrations.len(),
+        "Expected more than {} migration runs, but only found {total_runs}",
+        all_migrations.len()
+    );
+
+    let migrations = migration_manager::make_all_migrations(&clickhouse);
+    assert!(migration_manager::should_skip_migrations(&clickhouse, &migrations).await);
 }
 
 /// Migration 0013 has some checks that enforce that concurrent migrations can't break
@@ -799,7 +1049,10 @@ async fn test_concurrent_clickhouse_migrations() {
 #[tokio::test(flavor = "multi_thread")]
 async fn test_migration_0013_old_table() {
     let (clickhouse, _cleanup_db) = get_clean_clickhouse(false);
-    clickhouse.create_database().await.unwrap();
+    clickhouse
+        .create_database_and_migrations_table()
+        .await
+        .unwrap();
 
     // When creating a new migration, add it to the end of this array,
     // and adjust the call to `invoke_all!` to include the new array index.
@@ -834,13 +1087,19 @@ async fn test_migration_0013_old_table() {
     ];
 
     // Run migrations up to right before 0013
-    for migration in migrations.iter() {
-        migration_manager::run_migration(&clickhouse, migration.as_ref(), true)
-            .await
-            .unwrap();
+    for migration in migrations {
+        migration_manager::run_migration(RunMigrationArgs {
+            clickhouse: &clickhouse,
+            migration: migration.as_ref(),
+            clean_start: true,
+            manual_run: false,
+            is_replicated: false,
+        })
+        .await
+        .unwrap();
     }
     // Manually create a table that should not exist
-    let query = r#"
+    let query = r"
         CREATE TABLE IF NOT EXISTS InferenceById
         (
             id UUID, -- must be a UUIDv7
@@ -850,18 +1109,20 @@ async fn test_migration_0013_old_table() {
             function_type Enum('chat' = 1, 'json' = 2)
         ) ENGINE = MergeTree()
         ORDER BY id;
-    "#;
+    ";
     let _ = clickhouse
         .run_query_synchronous_no_params(query.to_string())
         .await
         .unwrap();
-    let err = migration_manager::run_migration(
-        &clickhouse,
-        &Migration0013 {
+    let err = migration_manager::run_migration(RunMigrationArgs {
+        clickhouse: &clickhouse,
+        migration: &Migration0013 {
             clickhouse: &clickhouse,
         },
-        false,
-    )
+        clean_start: false,
+        manual_run: false,
+        is_replicated: false,
+    })
     .await
     .unwrap_err();
     assert!(
@@ -877,7 +1138,10 @@ async fn test_migration_0013_old_table() {
 #[tokio::test(flavor = "multi_thread")]
 async fn test_migration_0013_data_no_table() {
     let (clickhouse, _cleanup_db) = get_clean_clickhouse(false);
-    clickhouse.create_database().await.unwrap();
+    clickhouse
+        .create_database_and_migrations_table()
+        .await
+        .unwrap();
 
     // When creating a new migration, add it to the end of this array,
     // and adjust the call to `invoke_all!` to include the new array index.
@@ -912,31 +1176,156 @@ async fn test_migration_0013_data_no_table() {
     ];
 
     // Run migrations up to right before 0013
-    for migration in migrations.iter() {
-        migration_manager::run_migration(&clickhouse, migration.as_ref(), true)
-            .await
-            .unwrap();
+    for migration in migrations {
+        migration_manager::run_migration(RunMigrationArgs {
+            clickhouse: &clickhouse,
+            migration: migration.as_ref(),
+            clean_start: true,
+            manual_run: false,
+            is_replicated: false,
+        })
+        .await
+        .unwrap();
     }
 
     // Add a row to the JsonInference table (would be very odd to have data in this table
     // but not an InferenceById table).
-    let query = r#"
+    let query = r"
         INSERT INTO JsonInference (id, function_name, variant_name, episode_id, input, output, output_schema, inference_params, processing_time_ms)
         VALUES (generateUUIDv7(), 'test_function', 'test_variant', generateUUIDv7(), 'input', 'output', 'output_schema', 'params', 100)
-    "#;
+    ";
     let _ = clickhouse
         .run_query_synchronous_no_params(query.to_string())
         .await
         .unwrap();
-    let err = migration_manager::run_migration(
-        &clickhouse,
-        &Migration0013 {
+    let err = migration_manager::run_migration(RunMigrationArgs {
+        clickhouse: &clickhouse,
+        migration: &Migration0013 {
             clickhouse: &clickhouse,
         },
-        false,
-    )
+        clean_start: false,
+        manual_run: false,
+        is_replicated: false,
+    })
     .await
     .unwrap_err();
     assert!(err.to_string()
         .contains("Data already exists in the ChatInference or JsonInference tables and InferenceById or InferenceByEpisodeId is missing. Please contact TensorZero team"));
 }
+
+#[tokio::test(flavor = "multi_thread")]
+#[traced_test]
+async fn test_run_migrations_clean() {
+    let (clickhouse, _cleanup_db) = get_clean_clickhouse(false);
+    migration_manager::run(RunMigrationManagerArgs {
+        clickhouse: &clickhouse,
+        manual_run: true,
+        skip_completed_migrations: true,
+    })
+    .await
+    .unwrap();
+    assert!(logs_contain("Database not found, assuming clean start"));
+    assert!(!logs_contain("All migrations have already been applied"));
+
+    // Run again, and we should skip all migrations
+    migration_manager::run(RunMigrationManagerArgs {
+        clickhouse: &clickhouse,
+        manual_run: true,
+        skip_completed_migrations: true,
+    })
+    .await
+    .unwrap();
+    assert!(logs_contain("All migrations have already been applied"));
+}
+
+#[tokio::test(flavor = "multi_thread")]
+#[traced_test]
+async fn test_run_migrations_fake_row() {
+    let (clickhouse, _cleanup_db) = get_clean_clickhouse(false);
+    clickhouse
+        .create_database_and_migrations_table()
+        .await
+        .unwrap();
+
+    struct Migration99999;
+
+    #[async_trait]
+    impl Migration for Migration99999 {
+        fn name(&self) -> String {
+            "Migration99999".to_string()
+        }
+        async fn can_apply(&self) -> Result<(), Error> {
+            Ok(())
+        }
+        async fn should_apply(&self) -> Result<bool, Error> {
+            Ok(true)
+        }
+        async fn apply(&self, _clean_start: bool) -> Result<(), Error> {
+            Ok(())
+        }
+        async fn has_succeeded(&self) -> Result<bool, Error> {
+            Ok(true)
+        }
+        fn rollback_instructions(&self) -> String {
+            "SELECT 1;".to_string()
+        }
+    }
+
+    let migration_manager_result = migration_manager::run(RunMigrationManagerArgs {
+        clickhouse: &clickhouse,
+        manual_run: false,
+        skip_completed_migrations: true,
+    })
+    .await;
+    if migration_manager_result.is_err() ^ clickhouse.is_cluster_configured() {
+        panic!("Migration manager should fail to run if and only if a cluster is configured");
+    }
+
+    // Run our fake migration to insert an unexpected row into `TensorZeroMigration`
+    // A subsequent normal run of migrations should *not* skip running migrations,
+    // since it will see an unexpected state
+    migration_manager::run_migration(RunMigrationArgs {
+        clickhouse: &clickhouse,
+        migration: &Migration99999,
+        manual_run: false,
+        clean_start: true,
+        is_replicated: false,
+    })
+    .await
+    .unwrap();
+
+    let migrations = migration_manager::make_all_migrations(&clickhouse);
+    assert!(!migration_manager::should_skip_migrations(&clickhouse, &migrations).await);
+
+    let migration_manager_result = migration_manager::run(RunMigrationManagerArgs {
+        clickhouse: &clickhouse,
+        skip_completed_migrations: true,
+        manual_run: false,
+    })
+    .await;
+    if migration_manager_result.is_err() ^ clickhouse.is_cluster_configured() {
+        panic!("Migration manager should fail to run if and only if a cluster is configured");
+    }
+    assert!(!logs_contain("already been applied"));
+
+    let rows = migration_manager::get_all_migration_records(&clickhouse)
+        .await
+        .unwrap();
+
+    let mut actual_migration_ids = rows.iter().map(|r| r.migration_id).collect::<Vec<_>>();
+    actual_migration_ids.sort();
+
+    let all_migrations = migration_manager::make_all_migrations(&clickhouse);
+    let mut expected_migration_ids = all_migrations
+        .iter()
+        .map(|m| m.migration_num().unwrap())
+        .collect::<Vec<_>>();
+    expected_migration_ids.push(99999);
+    expected_migration_ids.sort();
+
+    if clickhouse.is_cluster_configured() {
+        assert_eq!(actual_migration_ids, vec![99999]);
+    } else {
+        assert_eq!(actual_migration_ids, expected_migration_ids);
+    }
+}
