@@ -4,9 +4,11 @@ mod best_of_n;
 mod cache;
 mod clickhouse;
 mod common;
+mod config;
 mod datasets;
 mod dicl;
 mod dynamic_evaluations;
+mod dynamic_variants;
 mod fallback;
 mod feedback;
 mod health;
